# coding: utf-8

"""An :class:`.Enum` class with additional features."""

from __future__ import annotations
import logging
import os
from datetime import datetime

import numpy
from silx.io.url import DataUrl
from silx.io.utils import get_data
from silx.io.utils import open as open_hdf5
from silx.utils.deprecation import deprecated
from silx.utils.enum import Enum as _Enum
from tomoscan.esrf.scan.utils import cwd_context
from tomoscan.io import HDF5File

from nxtomo.nxobject.nxdetector import ImageKey
from nxtomo.utils.frameappender import FrameAppender
from nxtomo.application.nxtomo import NXtomo

try:
    import hdf5plugin  # noqa F401
except ImportError:
    pass
import uuid
from silx.io.utils import h5py_read_dataset

__all__ = [
    "embed_url",
    "FileExtension",
    "get_file_name",
    "get_tuple_of_keys_from_cmd",
    "is_nx_tomo_entry",
    "add_dark_flat_nx_file",
    "change_image_key_control",
    "str_datetime_to_numpy_datetime64",
    "strip_extension",
]


def embed_url(url: DataUrl, output_file: str) -> DataUrl:
    """
    Create a dataset under duplicate_data and with a random name
    to store it

    :param DataUrl url: dataset to be copied
    :param output_file: where to store the dataset
    :param expected_type: some metadata to put in copied dataset attributes
    :param data: data loaded from url is already loaded
    """
    if not isinstance(url, DataUrl):
        return url
    elif url.file_path() == output_file:
        return url
    else:
        embed_data_path = "/".join(("/duplicate_data", str(uuid.uuid1())))
        with cwd_context(os.path.dirname(os.path.abspath(output_file))):
            with HDF5File(output_file, "a") as h5s:
                h5s[embed_data_path] = get_data(url)
                h5s[embed_data_path].attrs["original_url"] = url.path()
            return DataUrl(
                file_path=output_file, data_path=embed_data_path, scheme="silx"
            )


class FileExtension(_Enum):
    H5 = ".h5"
    HDF5 = ".hdf5"
    NX = ".nx"
    NXS = ".nxs"


def get_file_name(file_name, extension, check=True):
    """
    set the given extension

    :param file_name: name of the file
    :param extension: extension to give
    :param check: if check, already check if the file as one of the
                       '_FileExtension'
    """
    if isinstance(extension, str):
        extension = FileExtension(extension.lower())
    assert isinstance(extension, FileExtension)
    if check:
        for item in FileExtension:
            if file_name.lower().endswith(item.value):
                return file_name
    return file_name + extension.value


def get_tuple_of_keys_from_cmd(cmd_value: str) -> tuple:
    """Return a tuple"""
    return tuple(cmd_value.split(","))


def is_nx_tomo_entry(file_path, entry):
    """

    :param file_path: hdf5 file path
    :param entry: entry to check
    :return: True if the entry is an NXTomo entry
    """
    if not os.path.exists(file_path):
        return False
    else:
        with open_hdf5(file_path) as h5s:
            if entry not in h5s:
                return False
            node = h5s[entry]
            return NXtomo.node_is_nxtomo(node)


def add_dark_flat_nx_file(
    file_path: str,
    entry: str,
    darks_start: numpy.ndarray | DataUrl | None = None,
    flats_start: numpy.ndarray | DataUrl | None = None,
    darks_end: numpy.ndarray | DataUrl | None = None,
    flats_end: numpy.ndarray | DataUrl | None = None,
    extras: dict | None = None,
    logger: None | logging.Logger = None,
    embed_data: bool = False,
):
    """
    This will get all data from entry@input_file and patch them with provided
    dark and / or flat(s).
    We consider the sequence as: dark, start_flat, projections, end_flat.

    Behavior regarding data type and target dataset:

    * if dataset at `entry` already exists:
        * if dataset at `entry` is a 'standard' dataset:
            * data will be loaded if necessary and `enrty` will be updated
        * if dataset at `entry` is a virtual dataset:
            * if `data` is a numpy array then we raise an error: the data should
              already be saved somewhere and you should provide a DataUrl
            * if `data` is a DataUrl then the virtual dataset is updated and
              a virtual source pointing to the
              DataUrl.file_path()@DataUrl.data_path() is added to the layout
    * if a new dataset `entry` need to be added:
        * if `data` is a numpy array then we create a new 'standard' Dataset
        * if `data` is a DataUrl then a new virtual dataset will be created

    note: Datasets `image_key`, `image_key_control`, `rotation_angle` and
    `count_time` will be copied each time.

    :param file_path: NXTomo file containing data to be patched
    :param entry: entry to be patched
    :param darks_start: (3D) numpy array containing the first dark serie if any
    :param flats_start: (3D) numpy array containing the first flat if any
    :param darks_end: (3D) numpy array containing dark the second dark serie if
                      any
    :param flats_end: (3D) numpy array containing the second flat if any
    :param extras: dictionary to specify some parameters for flats and dark
                   like rotation angle.
                   valid keys: 'start_dark', 'end_dark', 'start_flag',
                   'end_flag'.
                   Values should be a dictionary of 'NXTomo' keys with
                   values to be set instead of 'default values'.
                   Possible values are:
                   * `count_time`
                   * `rotation_angle`
    :param logger: object for logs
    :param embed_data: if True then each external data will be copy
                            under a 'duplicate_data' folder
    """
    if extras is None:
        extras = {}
    else:
        for key in extras:
            valid_extra_keys = ("darks_start", "darks_end", "flats_start", "flats_end")
            if key not in valid_extra_keys:
                raise ValueError(
                    f"{key} is not recognized. Valid values are {valid_extra_keys}"
                )

    if embed_data is True:
        darks_start = embed_url(darks_start, output_file=file_path)
        darks_end = embed_url(darks_end, output_file=file_path)
        flats_start = embed_url(flats_start, output_file=file_path)
        flats_end = embed_url(flats_end, output_file=file_path)
    else:
        for url in (darks_start, darks_end, flats_start, flats_end):
            if url is not None and isinstance(url, DataUrl):
                if isinstance(url.data_slice(), slice):
                    if url.data_slice().step not in (None, 1):
                        raise ValueError(
                            "When data is not embed slice `step`"
                            "must be None or 1. Other values are"
                            f"not handled. Failing url is {url}"
                        )

    # !!! warning: order of dark / flat treatments import
    data_names = "flats_start", "darks_end", "flats_end", "darks_start"
    datas = flats_start, darks_end, flats_end, darks_start
    keys_value = (
        ImageKey.FLAT_FIELD.value,
        ImageKey.DARK_FIELD.value,
        ImageKey.FLAT_FIELD.value,
        ImageKey.DARK_FIELD.value,
    )
    wheres = "start", "end", "end", "start"  # warning: order import

    for d_n, data, key, where in zip(data_names, datas, keys_value, wheres):
        if data is None:
            continue
        n_frames_to_insert = 1
        if isinstance(data, str):
            data = DataUrl(path=data)
        if isinstance(data, numpy.ndarray) and data.ndim == 3:
            n_frames_to_insert = data.shape[0]
        elif isinstance(data, DataUrl):
            with open_hdf5(data.file_path()) as h5s:
                if data.data_path() not in h5s:
                    raise KeyError(
                        f"Path given ({data.data_path()}) is not in {data.file_path}"
                    )
            data_node = get_data(data)
            if data_node.ndim == 3:
                n_frames_to_insert = data_node.shape[0]
        else:
            raise TypeError(f"{type(data)} as input is not managed")

        if logger is not None:
            logger.info(f"insert {type(data)} frame of type {key} at the {where}")
        # update 'data' dataset
        data_path = os.path.join(entry, "instrument", "detector", "data")
        FrameAppender(
            data, file_path, data_path=data_path, where=where, logger=logger
        ).process()
        # update image-key and image_key_control (we are not managing the
        # 'alignment projection here so values are identical')
        ik_path = os.path.join(entry, "instrument", "detector", "image_key")
        ikc_path = os.path.join(entry, "instrument", "detector", "image_key_control")
        for path in (ik_path, ikc_path):
            FrameAppender(
                [key] * n_frames_to_insert,
                file_path,
                data_path=path,
                where=where,
                logger=logger,
            ).process()

        # add 'other' necessaries key:
        count_time_path = os.path.join(
            entry,
            "instrument",
            "detector",
            "count_time",
        )
        rotation_angle_path = os.path.join(entry, "sample", "rotation_angle")
        x_translation_path = os.path.join(entry, "sample", "x_translation")
        translation_y_path = os.path.join(entry, "sample", "translation_y")
        translation_z_path = os.path.join(entry, "sample", "translation_z")
        control_data_path = os.path.join(entry, "control", "data")
        data_key_paths = (
            count_time_path,
            rotation_angle_path,
            x_translation_path,
            translation_y_path,
            translation_z_path,
            control_data_path,
        )
        mandatory_keys = (
            "count_time",
            "rotation_angle",
        )
        optional_keys = (
            "x_translation",
            "translation_y",
            "translation_z",
            "control/data",
        )

        data_keys = tuple(list(mandatory_keys) + list(optional_keys))

        for data_key, data_key_path in zip(data_keys, data_key_paths):
            data_to_insert = None
            if d_n in extras and data_key in extras[d_n]:
                provided_value = extras[d_n][data_key]
                if _is_iterable(provided_value):
                    if len(provided_value) != n_frames_to_insert:
                        raise ValueError(
                            "Given value to store from extras has"
                            f" incoherent length({len(provided_value)}) compare to "
                            f"the number of frame to save ({n_frames_to_insert})"
                        )
                    else:
                        data_to_insert = provided_value
                else:
                    try:
                        data_to_insert = [provided_value] * n_frames_to_insert
                    except Exception as e:
                        logger.error(f"Fail to create data to insert. Error is {e}")
                        return
            else:
                # get default values
                def get_default_value(location, where_):
                    with open_hdf5(file_path) as h5s:
                        if location not in h5s:
                            return None
                        existing_data = h5s[location]
                        if where_ == "start":
                            return existing_data[0]
                        else:
                            return existing_data[-1]

                try:
                    default_value = get_default_value(data_key_path, where)
                except Exception:
                    default_value = None
                if default_value is None:
                    msg = f"Unable to define a default value for {data_key_path}. Location empty in {file_path}"
                    if data_key in mandatory_keys:
                        raise ValueError(msg)
                    elif logger:
                        logger.warning(msg)
                    continue
                elif logger:
                    logger.debug(
                        f"No value(s) provided for {data_key_path}. Extract some default value ({default_value})."
                    )
                data_to_insert = [default_value] * n_frames_to_insert

            if data_to_insert is not None:
                FrameAppender(
                    data_to_insert,
                    file_path,
                    data_path=data_key_path,
                    where=where,
                    logger=logger,
                ).process()


@deprecated(replacement="_FrameAppender", since_version="0.5.0")
def _insert_frame_data(data, file_path, data_path, where, logger=None):
    """
    This function is used to insert some frame(s) (numpy 2D or 3D to an
    existing dataset. Before the existing array or After.

    :param data:
    :param file_path:
    :param data_path: If the path point to a virtual dataset them this one
                      will be updated but data should be a DataUrl. Of the
                      same shape. Else we will update the data_path by
                      extending the dataset.
    :param where:
    :raises TypeError: In the case the data type and existing data_path are
                       incompatible.
    """
    fa = FrameAppender(
        data=data, file_path=file_path, data_path=data_path, where=where, logger=logger
    )
    return fa.process()


def change_image_key_control(
    file_path: str,
    entry: str,
    frames_indexes,
    image_key_control_value: int | ImageKey,
    logger=None,
):
    """
    Will modify image_key and image_key_control values for the requested
    frames.

    :param file_path: path the nexus file
    :param entry: name of the entry to modify
    :param frames_indexes: index of the frame for which we want to modify
                           the image key
    :param image_key_control_value:
    :param logging.Logger logger: logger
    """
    if not isinstance(frames_indexes, slice) and not _is_iterable(frames_indexes):
        raise TypeError("`frame_indexes` should be a sequence or slice")
    if logger:
        logger.info(
            "Update frames {frames_indexes} to"
            "{image_key_control_value} of {entry}@{file_path}"
            "".format(
                frames_indexes=frames_indexes,
                image_key_control_value=image_key_control_value,
                entry=entry,
                file_path=file_path,
            )
        )

    image_key_control_value = ImageKey(image_key_control_value)
    with HDF5File(file_path, mode="a") as h5s:
        node = h5s[entry]
        image_keys_path = "/".join(("instrument", "detector", "image_key"))
        image_keys = h5py_read_dataset(node[image_keys_path])
        image_keys_control_path = "/".join(
            ("instrument", "detector", "image_key_control")
        )
        image_keys_control = h5py_read_dataset(node[image_keys_control_path])
        # filter frame indexes
        if isinstance(frames_indexes, slice):
            step = frames_indexes.step
            if step is None:
                step = 1
            stop = frames_indexes.stop
            if stop in (None, -1):
                stop = len(image_keys)
            frames_indexes = list(range(frames_indexes.start, stop, step))
        frames_indexes = list(
            filter(lambda x: 0 <= x <= len(image_keys_control), frames_indexes)
        )
        # manage image_key_control
        image_keys_control[frames_indexes] = image_key_control_value.value
        node[image_keys_control_path][:] = image_keys_control
        # manage image_key. In this case we should get rid of Alignment values
        # and replace it by Projection values
        image_key_value = image_key_control_value
        if image_key_value is ImageKey.ALIGNMENT:
            image_key_value = ImageKey.PROJECTION
        image_keys[frames_indexes] = image_key_value.value
        node[image_keys_path][:] = image_keys


def str_datetime_to_numpy_datetime64(my_datetime: str | datetime) -> numpy.datetime64:
    # numpy deprecates time zone awarness conversion to numpy.datetime64.
    # so we remove the time zone info.
    if isinstance(my_datetime, str):
        datetime_as_datetime = datetime.fromisoformat(my_datetime)
    elif isinstance(my_datetime, datetime):
        datetime_as_datetime = my_datetime
    else:
        raise TypeError(
            f"my_datetime is expected to be a str or an instance of datetime. Not {type(my_datetime)}"
        )

    datetime_as_utc_datetime = datetime_as_datetime.astimezone(None)
    tz_free_datetime_as_datetime = datetime_as_utc_datetime.replace(tzinfo=None)
    return numpy.datetime64(tz_free_datetime_as_datetime).astype("<M8[ms]")


def strip_extension(filename, logger=None):
    if filename.endswith((".nx", ".h5")):
        return filename[:-3]
    elif filename.endswith(".hdf5"):
        return filename[:-5]
    else:
        if logger is not None:
            logger.warning(f"Unusual file name {filename} has no known postfix")
        return filename


def _is_iterable(value):
    if isinstance(value, (str, bytes)):
        return False
    try:
        iter(value)
    except TypeError:
        return False
    return True



"""module to get NXtomo versionning as it can evolve with time"""

LATEST_VERSION = 1.2

CURRENT_OUTPUT_VERSION = LATEST_VERSION


def version():
    return CURRENT_OUTPUT_VERSION


# Information regarding Format
# Format 1.0
#    NXtomo entry (one per acquisition)
#       |-> beam
#             |-> incident energy (optional 0D)
#       |-> instrument (NXinstrument)
#             |-> detector (NXdetector)
#                  |-> count_time (optional 1D dataset)
#                  |-> data (mandatory 3D dataset)
#                  |-> distance (optional 0D dataset)
#                  |-> field_of_fiew (optional str)
#                  |-> image_key (madatory 1D dataset)
#                  |-> image_key_control (optional 1D dataset)
#                  |-> x_pixel_size (float)
#                  |-> y_pixel_size (float)
#       |-> sample (NXsample)
#             |-> name (optional)
#             |-> rotation_angle - mandatory (1D dataset in degree)
#             |-> x_translation - optional (1D dataset)
#             |-> y_translation - optional (1D dataset)
#             |-> z_translation - optional (1D dataset)
#
# Format 1.1:
#    * move beam to the NXinstrument
#       * Keep compatibility by providing a link to beam at the root level
#    * add optional dataset instrument/name
#    * sample/name is moved to title
#    * sample/sample_name is moved to sample/name
#    * add NXsource under instrument:
#       |-> instrument (NXinstrument)
#              |-> source (optional NXsource)
#                   |-> name (optional)
#                   |-> type (optional)



# coding: utf-8

"""
Application to patch a NXtomo entry. invalidating some frame or adding some.

.. program-output:: patch-nx --help

"""

import argparse
import logging
import os

from silx.io.url import DataUrl
from silx.utils.enum import Enum as _Enum
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan

from nxtomomill import utils
from nxtomo.nxobject.nxdetector import ImageKey

logging.basicConfig(level=logging.DEBUG)
_logger = logging.getLogger(__name__)


_SILX_DATA_URL = "www.silx.org/doc/silx/latest/modules/io/url.html?highlight=dataurl#silx.io.url.DataUrl"


_INFO_URL = (
    'url should be providing the "default?silx way": '
    "silx:///data/image.edf?path=/scan_0/detector/data "
    f"(see {_SILX_DATA_URL})"
    "of just by giving dataset_path@file_path"
)


class _ImageKeyName(_Enum):
    ALIGNMENT = "alignment"
    PROJECTION = "projection"
    FLAT_FIELD = "flat"
    DARK_FIELD = "dark"
    INVALID = "invalid"

    @staticmethod
    def to_image_key(image_key) -> ImageKey:
        image_key = _ImageKeyName(image_key.lower())
        if image_key is _ImageKeyName.ALIGNMENT:
            return ImageKey.ALIGNMENT
        elif image_key is _ImageKeyName.PROJECTION:
            return ImageKey.PROJECTION
        elif image_key is _ImageKeyName.DARK_FIELD:
            return ImageKey.DARK_FIELD
        elif image_key is _ImageKeyName.FLAT_FIELD:
            return ImageKey.FLAT_FIELD
        elif image_key is _ImageKeyName.INVALID:
            return ImageKey.INVALID
        else:
            raise ValueError(f"{image_key} not handled")


_INFO_FRAME_INPUT = (
    "Frames can be provided three ways: \n"
    "- as a list: frame_index_1,frame_index_2\n"
    "- as a python slice: from:to:step\n"
    f"- as an image key value. Valid values are {[item.value for item in _ImageKeyName]}\n"
)


def _extract_data_url(url_as_a_str):
    """
    Extract url from a string
    """
    if url_as_a_str is None:
        return None
    elif "@" in url_as_a_str:
        try:
            entry, file_path = url_as_a_str.split("@")
        except Exception:
            _logger.error(f"Fail to create an url from {url_as_a_str}. {_INFO_URL}")
            return None
        else:
            url = DataUrl(file_path=file_path, data_path=entry, scheme="silx")
            return url
    else:
        try:
            url = DataUrl(path=url_as_a_str)
        except Exception as e:
            _logger.error(
                f"Fail to create an url from {url_as_a_str}."
                f"Reason is {e}. For more information see {_SILX_DATA_URL}"
            )
            return None
        else:
            return url


def _get_slice_to_modify(slice_as_str, master_file, entry):
    """
    Return a list of int or a `slice` from  slice_as_str
    :param slice_as_str:
    :return: slice to be modify on the image_key and image_key_control dataset
    """
    if slice_as_str is None:
        return None
    elif slice_as_str.lower() in [item.value for item in _ImageKeyName]:
        image_key = _ImageKeyName.to_image_key(slice_as_str)
        scan = NXtomoScan(master_file, entry)
        frames = scan.frames
        slices = []
        for frame in frames:
            if frame.image_key is image_key:
                slices.append(frame.index)
        return slices
    elif ":" in slice_as_str:
        elmts = slice_as_str.split(":")

        def get_value(index):
            if index >= len(elmts):
                return None
            elif elmts[index] == "":
                return None
            else:
                return int(elmts[index])

        from_ = get_value(0)
        to_ = get_value(1)
        step = get_value(2)
        return slice(from_, to_, step)
    else:
        return [int(index) for index in slice_as_str.split(",")]


def main(argv):
    """ """
    parser = argparse.ArgumentParser(
        description="Insert dark and / or flat frames and metadata into an"
        "existing NXTomo file from url(s)."
    )
    parser.add_argument("file_path", help="NXTomo file to patch")
    parser.add_argument("entry", help="entry in the provided file")

    # dark and flat options
    parser.add_argument(
        "--darks-at-start",
        "--darks-start",
        default=None,
        help="url to the dataset containing darks to be store at"
        "the beginning. " + _INFO_URL,
    )
    parser.add_argument(
        "--darks-at-end",
        "--darks-end",
        default=None,
        help="url to the dataset containing darks to be store at"
        "the end." + _INFO_URL,
    )
    parser.add_argument(
        "--flats-at-start",
        "--flats-start",
        default=None,
        help="url to the dataset containing flats to be store at"
        "the beginning of the acquisition sequence (made before "
        "projections acquisition). " + _INFO_URL,
    )
    parser.add_argument(
        "--flats-at-end",
        "--flats-end",
        default=None,
        help="url to the dataset containing flats to be store at"
        "the beginning of end of the sequence (made before "
        "projections acquisition). " + _INFO_URL,
    )

    # modify frame type option
    parser.add_argument(
        "--invalid-frames",
        default=None,
        help="Define the set of frames to be mark as invalid. " + _INFO_FRAME_INPUT,
    )
    parser.add_argument(
        "--update-to-projection",
        "--update-to-proj",
        default=None,
        help="Define the set of frames to be mark as projection. "
        "" + _INFO_FRAME_INPUT,
    )
    parser.add_argument(
        "--update-to-dark",
        default=None,
        help="Define the set of frames to be mark as dark. " + _INFO_FRAME_INPUT,
    )
    parser.add_argument(
        "--update-to-flat",
        default=None,
        help="Define the set of frames to be mark as flat. " + _INFO_FRAME_INPUT,
    )
    parser.add_argument(
        "--update-to-alignment",
        default=None,
        help="Define the set of frames to be mark as alignment. "
        "" + _INFO_FRAME_INPUT,
    )
    parser.add_argument(
        "--embed-data",
        default=False,
        action="store_true",
        help="Embed data from url in the file if not already inside",
    )

    options = parser.parse_args(argv[1:])

    # get information for adding dark / flat
    darks_start_url = _extract_data_url(options.darks_at_start)
    darks_end_url = _extract_data_url(options.darks_at_end)
    flat_start_url = _extract_data_url(options.flats_at_start)
    flat_end_url = _extract_data_url(options.flats_at_end)

    patch_det_data = darks_start_url or flat_start_url or flat_end_url or darks_end_url

    # get information for modifying image_key
    slice_to_update_to_dark = _get_slice_to_modify(
        options.update_to_dark, master_file=options.file_path, entry=options.entry
    )
    slice_to_update_to_flat = _get_slice_to_modify(
        options.update_to_flat, master_file=options.file_path, entry=options.entry
    )
    slice_to_update_to_projection = _get_slice_to_modify(
        options.update_to_projection, master_file=options.file_path, entry=options.entry
    )
    slice_to_update_to_alignment = _get_slice_to_modify(
        options.update_to_alignment, master_file=options.file_path, entry=options.entry
    )
    slice_to_invalid = _get_slice_to_modify(
        options.invalid_frames, master_file=options.file_path, entry=options.entry
    )
    patch_image_key = (
        slice_to_update_to_dark
        or slice_to_update_to_flat
        or slice_to_update_to_projection
        or slice_to_update_to_alignment
        or slice_to_invalid
    )

    if patch_det_data and patch_image_key:
        _logger.info(
            "Both adding dark / flat and modifying `image_key` / "
            "`image_key_control` are requested. Will first add "
            "dark / flat then modify `image_key` / "
            "`image_key_control`."
        )

    elif not (patch_det_data or patch_image_key):
        _logger.warning(
            "No url provided for dark or flats or frame type to "
            "modify. Nothing to be done."
        )
    elif not utils.is_nx_tomo_entry(file_path=options.file_path, entry=options.entry):
        _logger.error(
            f"{options.entry}@{options.file_path} is not recognized as a valid NXTomo entry."
        )
    elif not os.access(options.file_path, os.W_OK):
        _logger.error(f"You don't have rights to write on {options.file_path}.")
    else:
        slices_patch = {
            ImageKey.ALIGNMENT: slice_to_update_to_alignment,
            ImageKey.PROJECTION: slice_to_update_to_projection,
            ImageKey.FLAT_FIELD: slice_to_update_to_flat,
            ImageKey.DARK_FIELD: slice_to_update_to_dark,
            ImageKey.INVALID: slice_to_invalid,
        }
        if patch_image_key:
            _logger.info("start updating frames")

        for image_key_type, frames_to_update in slices_patch.items():
            if frames_to_update is None:
                continue
            utils.change_image_key_control(
                file_path=options.file_path,
                entry=options.entry,
                frames_indexes=frames_to_update,
                image_key_control_value=image_key_type,
                logger=_logger,
            )

        if patch_det_data:
            _logger.info("start adding dark and flat field")
            utils.add_dark_flat_nx_file(
                file_path=options.file_path,
                entry=options.entry,
                darks_start=darks_start_url,
                flats_start=flat_start_url,
                darks_end=darks_end_url,
                flats_end=flat_end_url,
                embed_data=True,
                logger=_logger,
            )



# coding: utf-8

"""
Application to convert from dx file (HDF5) to NXTomo (HDF5) file

.. program-output:: nxtomomill dxfile2nx --help

For a complete tutorial you can have a look at :ref:`dxfile2nxtutorial`

"""

import argparse
import logging
import pint

from nxtomomill import utils
from nxtomomill.converter.dxfile.dxfileconverter import from_dx_to_nx

_logger = logging.getLogger(__name__)
_ureg = pint.get_application_registry()


def convert_2elmts__tuple_to_float(input_str) -> tuple:
    if input_str == (None, None):
        return input_str

    try:
        tmp_str = input_str.replace(" ", "")
        tmp_str = tmp_str.replace(";", ",")
        if tmp_str.startswith("("):
            tmp_str = tmp_str[1:]
        if tmp_str.startswith(")"):
            tmp_str = tmp_str[:-1]
        v1, v2 = tmp_str.split(",")
    except Exception as e:
        raise ValueError(
            f"Unable to convert {input_str} to a tuple of two float. Reason is {e}"
        )
    else:
        return float(v1), float(v2)


def _get_pixel_size(input_str) -> tuple:
    try:
        values = convert_2elmts__tuple_to_float(input_str)
    except ValueError:
        try:
            value = float(input_str)
        except Exception:
            raise ValueError(
                f"Unable to convert {input_str} to pixel size."
                f"Should be provided as a tuple (sample x_pixel_size, sample y_pixel_size) "
                f"or as a single value pixel_size"
            )
        else:
            return value, value
    else:
        return values


def main(argv):
    """ """
    parser = argparse.ArgumentParser(
        description="convert acquisition contained in provided dx file to "
        "NXTomo entry"
    )
    parser.add_argument("input_file", help="master file of the " "acquisition")
    parser.add_argument("output_file", help="output .nx or .h5 file", default=None)
    # usually it looks like the dxfile contains only one acquisition at the
    # root level
    parser.add_argument(
        "--input_entry",
        help="h5py group path to be converted",
        default="/",
    )
    parser.add_argument(
        "--output_entry",
        help="h5py group path to store the NXTomo",
        default="entry0000",
    )
    parser.add_argument(
        "--file_extension",
        default=".nx",
        help="extension of the output file. Valid values are "
        "" + "/".join([item.value for item in utils.FileExtension]),
    )
    # as scan range or rotation angle are not stored in dxfile we
    # provide the default value which is the "standard" case
    parser.add_argument(
        "--scan-range",
        default="0,180",
        help="scan range of the projections. Dark and flat will always be "
        "considered with rotation angle == scan_range[0]",
    )
    # as pixel sizes are not stored in the dxfile and requested for paganin
    # we provide some default value
    parser.add_argument(
        "--pixel-size",
        default=(None, None),
        help="pixel size in meter as (x pixel size, y pixel size) or as a single value",
    )
    parser.add_argument(
        "--fov",
        "--field-of-view",
        default=None,
        help="field of view. Can be Full or Half",
        dest="field_of_view",
    )
    # as distance is not stored in the dxfile and requested for paganin
    # we provide some default value
    parser.add_argument(
        "--distance",
        "--detector-sample-distance",
        default=None,
        help="sample to detector distance (in meter)",
        dest="distance",
    )
    parser.add_argument(
        "--energy",
        "--incident-beam-energy",
        default=None,
        help="incident beam energy in keV",
        dest="energy",
    )
    parser.add_argument(
        "--overwrite",
        help="Do not ask for user permission to overwrite output files",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--data-copy",
        help="Force data duplication of frames. This will permit to have an "
        "'all-embed' file. Otherwise we will have link between the dxfile "
        "and the NXTomo.",
        action="store_true",
        dest="duplicate_data",
        default=False,
    )
    options = parser.parse_args(argv[1:])

    distance = options.distance
    if distance is not None:
        distance = float(options.distance)
    energy = options.energy
    if energy is not None:
        energy = float(options.energy) * _ureg.keV
    if options.duplicate_data is False:
        _logger.warning(
            "Generated file will contain relative links to the "
            "(dxfile) source file. You should insure the "
            "relative position of the input file and the output "
            "stay constant to protect links. (use --copy-data) "
            "option to avoid links"
        )

    from_dx_to_nx(
        input_file=options.input_file,
        output_file=options.output_file,
        file_extension=options.file_extension,
        input_entry=options.input_entry,
        output_entry=options.output_entry,
        scan_range=convert_2elmts__tuple_to_float(options.scan_range),
        pixel_size=_get_pixel_size(options.pixel_size),
        field_of_view=options.field_of_view,
        distance=distance,
        overwrite=options.overwrite,
        duplicate_data=options.duplicate_data,
        energy=energy,
    )



from __future__ import annotations


def split_timestamps(my_array, n_part: int):
    """
    split given array into n_part (as equal as possible)
    :param sequence my_array:
    """
    array_size = len(my_array)
    if array_size < n_part:
        yield my_array
    else:
        start = 0
        for _ in range(n_part):
            end = max(start + int(array_size / n_part) + 1, array_size)
            yield my_array[start:end]
            start = end



"""Scan dedicated for bliss format - based on EDF files"""

from __future__ import annotations

import logging
import os
import glob
import pint

import numpy

from tomoscan.identifier import ScanIdentifier
from tomoscan.scanbase import TomoScanBase
from tomoscan.utils import docstring


from dataclasses import dataclass, field
import numpy as np
from numpy.typing import NDArray, DTypeLike
from silx.io.utils import h5py_read_dataset

try:
    from tqdm.auto import tqdm
except ImportError:
    has_tqdm = False
else:
    has_tqdm = True
import h5py

_ureg = pint.get_application_registry()
_logger = logging.getLogger(__name__)

try:
    import tifffile  # noqa #F401 needed for later possible lazy loading
except ImportError:
    has_tifffile = False
else:
    has_tifffile = True

__all__ = ["FluoTomoScanBase", "FluoTomoScan3D", "FluoTomoScan2D"]


@dataclass
class FluoTomoScanBase:
    """Dataset manipulation class."""

    scan: str
    dataset_basename: str
    detectors: tuple = ()

    dtype: DTypeLike = numpy.float32
    verbose: bool = False

    angles: float | None = None
    """rotation angles in degree"""
    el_lines: dict[str, list[str]] = field(default_factory=dict)
    pixel_size: float | None = None
    """pixel size in meter"""
    energy: float | None = None
    """energy in keV"""
    detected_folders: list[str] = field(default_factory=list)

    def __post_init__(self):
        self.detected_folders = self.detect_folders()
        self.detected_detectors = tuple(self.detect_detectors())
        self.get_metadata_from_h5_file()
        _logger.info(f"Detectors: {self.detected_detectors}")
        if len(self.detectors) == 0:
            self.detectors = self.detected_detectors
        for det in self.detectors:
            if det not in self.detected_detectors:
                raise ValueError(
                    f"The detector should be in {self.detected_detectors} and is {self.detectors}"
                )
        self.detect_elements()

        if self.angles is None:
            self.angles = self.detect_rot_angles()

    def detect_folders(self) -> list[str]:
        """List all folders to process."""
        raise NotImplementedError("Base class")

    @property
    def rot_angles_deg(self) -> NDArray:
        if self.angles is None:
            raise ValueError("Rotation angles not initialized")

        return self.angles * _ureg.deg

    @property
    def rot_angles_rad(self) -> NDArray:
        return self.rot_angles_deg.to(_ureg.rad)

    def detect_rot_angles(self) -> NDArray:
        """Build rotation angles list."""
        raise NotImplementedError("Base class")

    def _check_ready_to_load_data(self, det):
        if not has_tifffile:
            raise RuntimeError("tiffile not install. Cannot load data.")
        if det not in self.detectors:
            raise RuntimeError(
                f"The detector {det} is invalid. Valid ones are {self.detectors}"
            )
        if self.angles is None:
            raise RuntimeError("Rotation angles not initialized")

        if self.detectors is None:
            raise RuntimeError("Detectors not initialized")

    def get_metadata_from_h5_file(self):
        if len(self.detected_folders) == 0:
            raise ValueError("No folder found, unable to load metadata")
        h5_path = os.path.join(self.scan, self.detected_folders[0])
        h5_files = glob.glob1(h5_path, "*.h5")
        if len(h5_files) > 1:
            raise ValueError(
                "More than one hdf5 file in scan directory. Expect only ONE to pick pixel size."
            )
        elif len(h5_files) == 0:
            pattern = os.path.join(h5_path, "*.h5")
            raise ValueError(
                f"Unable to find the hdf5 file in scan directory to pick pixel size. RegExp used is {pattern}"
            )
        else:
            with h5py.File(os.path.join(h5_path, h5_files[0]), "r") as f:
                if len(list(f.keys())) != 1:
                    raise ValueError(
                        f"H5 file should contain only one entry, found {len(list(f.keys()))}"
                    )
                else:
                    entry_name = list(f.keys())[0]
                    self.pixel_size = (
                        (
                            h5py_read_dataset(f[entry_name]["FLUO"]["pixelSize"])
                            * _ureg.micrometer
                        )
                        .to(_ureg.meter)
                        .magnitude
                    )
                    self.energy = float(
                        h5py_read_dataset(
                            f[entry_name]["instrument"]["monochromator"]["energy"]
                        )
                    )

    def detect_detectors(self):
        if len(self.detected_folders) == 0:
            raise ValueError("No folder found, unable to detect detectors")
        proj_1_dir = os.path.join(self.scan, "fluofit", self.detected_folders[0])
        detected_detectors = []
        file_names = glob.glob1(proj_1_dir, "IMG_*area_density_ngmm2.tif")
        for file in file_names:
            det = file.split("_")[1]
            if det not in detected_detectors:
                detected_detectors.append(det)
        if "" in detected_detectors:
            raise ValueError(
                f"Suspicious detector! Detected detectors are {detected_detectors}. Please use --detectors <det1>... where det1 are valid detector name."
            )

        return detected_detectors

    def detect_elements(self):
        if len(self.detected_folders) == 0:
            raise ValueError("No folder found, unable to detect elements")
        proj_1_dir = os.path.join(self.scan, "fluofit", self.detected_folders[0])
        detector = self.detectors[0]
        file_names = glob.glob1(proj_1_dir, f"IMG_{detector}*area_density_ngmm2.tif")
        for file in file_names:
            el_str = file.split("_")[2]
            element, line = el_str.split("-")
            try:
                if line not in self.el_lines[element]:
                    self.el_lines[element].append(line)
                    self.el_lines[element] = sorted(self.el_lines[element])
            except KeyError:
                self.el_lines[element] = [line]

    def load_data(self, det: str, element: str, line_ind: int = 0) -> NDArray:
        """Main function of class to load data."""
        raise NotImplementedError("Base class")

    @staticmethod
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        raise NotImplementedError("Not implemented for fluo-tomo yet.")

    @docstring(TomoScanBase)
    def get_identifier(self) -> ScanIdentifier:
        raise NotImplementedError("Not implemented for fluo-tomo yet.")


@dataclass
class FluoTomoScan3D(FluoTomoScanBase):
    """Dataset manipulation class."""

    def detect_folders(self):
        fit_folders = glob.glob1(
            os.path.join(self.scan, "fluofit"), rf"{self.dataset_basename}_projection*"
        )
        fit_folders.sort()
        proj_indices = [int(f[-3:]) for f in fit_folders]

        if len(fit_folders) < proj_indices[-1]:
            _logger.debug(
                f"The number of projections ({len(fit_folders)}) is lower than the largest projection index {proj_indices[-1]}. Some projections are probably missing."
            )

        if len(fit_folders) == 0:
            raise FileNotFoundError(
                "No projection was found in the fluofit folder. The searched for pattern is <scan_dir>/fluofit/<dataset_basename>_projection*'."
            )
        elif not os.path.isdir(os.path.join(self.scan, fit_folders[0])):
            raise FileNotFoundError(
                "Found fitted data folders but not the corresponding raw data folder."
            )
        else:
            return fit_folders

    def detect_rot_angles(self) -> NDArray:
        tmp_angles = []
        for f in self.detected_folders:
            prj_dir = os.path.join(self.scan, "fluofit", f)
            info_file = os.path.join(prj_dir, "info.txt")
            try:
                with open(info_file, "r") as f:
                    info_str = f.read()
                    tmp_angles.append(float(info_str.split(" ")[2]))
            except FileNotFoundError:
                _logger.debug(
                    f"{info_file} doesn't exist, while expected to be present in each projection folder."
                )
                raise FileNotFoundError(
                    f"{info_file} doesn't exist, while expected to be present in each projection folder."
                )
            _logger.info(f"Correctly found all {len(tmp_angles)} rotation angles.")
        return numpy.array(tmp_angles, ndmin=1, dtype=numpy.float32)

    def load_data(self, det: str, element: str, line_ind: int = 0) -> NDArray:

        self._check_ready_to_load_data(det)

        line = self.el_lines[element][line_ind]

        data_det = []

        description = f"Loading images of {element}-{line} ({det}): "

        if has_tqdm:
            proj_iterator = tqdm(
                range(len(self.detected_folders)),
                disable=self.verbose,
                desc=description,
            )
        else:
            proj_iterator = range(len(self.detected_folders))

        for ii_i in proj_iterator:
            proj_dir = os.path.join(
                self.scan,
                "fluofit",
                self.detected_folders[ii_i],
            )
            img_path = os.path.join(
                proj_dir, f"IMG_{det}_{element}-{line}_area_density_ngmm2.tif"
            )

            if self.verbose:
                _logger.info(f"Loading {ii_i + 1} / {len(self.angles)}: {img_path}")

            img = tifffile.imread(img_path)
            data_det.append(numpy.nan_to_num(numpy.array(img, dtype=self.dtype)))

        data = numpy.array(data_det)
        return numpy.ascontiguousarray(data)


@dataclass
class FluoTomoScan2D(FluoTomoScanBase):
    """Dataset manipulation class."""

    def get_metadata_from_h5_file(self):
        super().get_metadata_from_h5_file()
        h5_path = os.path.join(self.scan, self.detected_folders[0])
        h5_files = glob.glob1(h5_path, "*.h5")
        if len(h5_files) > 1:
            raise ValueError(
                "More than one hdf5 file in scan directory. Expect only ONE to pick pixel size."
            )
        elif len(h5_files) == 0:
            pattern = os.path.join(h5_path, "*.h5")
            raise ValueError(
                f"Unable to find the hdf5 file in scan directory to pick pixel size. RegExp used is {pattern}"
            )
        else:
            with h5py.File(os.path.join(h5_path, h5_files[0]), "r") as f:
                if len(list(f.keys())) != 1:
                    raise ValueError(
                        f"H5 file should contain only one entry, found {len(list(f.keys()))}"
                    )
                else:
                    entry_name = list(f.keys())[0]
                    self.scanRange_2 = float(
                        h5py_read_dataset(f[entry_name]["FLUO"]["scanRange_2"])
                    )
                    self.scanDim_2 = int(
                        h5py_read_dataset(f[entry_name]["FLUO"]["scanDim_2"])
                    )

    def detect_rot_angles(self) -> NDArray:
        nb_projs = self.scanDim_2
        angular_coverage = self.scanRange_2
        return np.linspace(0, angular_coverage, nb_projs, endpoint=True)

    def detect_folders(self):
        fit_folder = os.path.join(self.scan, "fluofit", self.dataset_basename)

        if not os.path.isdir(fit_folder):
            raise FileNotFoundError(
                f"No folder {fit_folder} was found in the fluofit folder. The searched for pattern is <scan_dir>/fluofit/<dataset_basename>'."
            )
        elif not os.path.isdir(os.path.join(self.scan, self.dataset_basename)):
            raise FileNotFoundError(
                "Found fitted data folders but not the corresponding raw data folder."
            )
        else:
            return [
                self.dataset_basename,
            ]

    def load_data(self, det: str, element: str, line_ind: int = 0) -> NDArray:
        self._check_ready_to_load_data(det)

        line = self.el_lines[element][line_ind]

        data_det = []

        description = f"Loading images of {element}-{line} ({det}): "

        if has_tqdm:
            slice_iterator = tqdm(
                range(len(self.detected_folders)),
                disable=self.verbose,
                desc=description,
            )
        else:
            slice_iterator = range(len(self.detected_folders))

        for ii_i in slice_iterator:

            proj_dir = os.path.join(
                self.scan,
                "fluofit",
                self.dataset_basename,  # WARNING: dataset_basename is ONE SINGLE sinogram.
            )
            img_path = os.path.join(
                proj_dir, f"IMG_{det}_{element}-{line}_area_density_ngmm2.tif"
            )

            if self.verbose:
                _logger.info(f"Loading {ii_i + 1} / {len(self.angles)}: {img_path}")

            img = tifffile.imread(img_path)
            data_det.append(numpy.nan_to_num(numpy.array(img, dtype=self.dtype)))

        data = numpy.array(data_det)
        return numpy.ascontiguousarray(data)



import os
import numpy
from fabio.edfimage import EdfImage
from silx.io.url import DataUrl
from nxtomomill.converter.edf.checks import compare_volumes
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from tomoscan.io import HDF5File


def test_compare_volumes(tmp_path):
    """test the compare_volumes function"""

    input_folder = tmp_path / "input"
    input_folder.mkdir()
    output_folder = tmp_path / "output"
    output_folder.mkdir()

    frame_1 = (
        numpy.arange(10, 110, dtype=numpy.float32)
        .reshape(1, 10, 10)
        .astype(numpy.float32)
    )
    frame_2 = (
        numpy.arange(30, 130, dtype=numpy.float32)
        .reshape(1, 10, 10)
        .astype(numpy.float32)
    )
    frame_3 = (
        numpy.arange(100, 200, dtype=numpy.float32)
        .reshape(1, 10, 10)
        .astype(numpy.float32)
    )

    edf_frame_files = []
    edf_frame_urls = []
    for i_frame, frame in enumerate((frame_1, frame_2, frame_3)):
        edf_writer = EdfImage(
            data=frame[0],  # [0]: easy way to concatenate the frame later
            header={},
        )
        edf_frame_file = os.path.join(input_folder, f"frame_{i_frame}.edf")
        edf_writer.write(edf_frame_file)
        edf_frame_files.append(edf_frame_file)
        edf_frame_urls.append(
            [
                DataUrl(file_path=edf_frame_file, scheme="fabio"),
                False,
            ]
        )

    scan_0_file = os.path.join(output_folder, "nx_tomo.nx")
    with HDF5File(scan_0_file, mode="w") as h5f:
        h5f["entry/instrument/detector/data"] = numpy.vstack(
            [frame_1, frame_2, frame_3]
        )

    assert len(compare_volumes(edf_frame_urls, NXtomoScan(scan_0_file, "entry"))) == 0

    # test inverting a frame
    scan_1_file = os.path.join(output_folder, "toto_tomo.nx")
    with HDF5File(scan_1_file, mode="w") as h5f:
        h5f["entry/instrument/detector/data"] = numpy.vstack(
            [frame_1, frame_1, frame_3]
        )

    assert len(compare_volumes(edf_frame_urls, NXtomoScan(scan_1_file, "entry"))) == 1
    # then ignore second frame
    edf_frame_urls[1][1] = True
    assert len(compare_volumes(edf_frame_urls, NXtomoScan(scan_1_file, "entry"))) == 0

    # test modifying the type
    scan_2_file = os.path.join(output_folder, "nx_tomo32.nx")
    with HDF5File(scan_2_file, mode="w") as h5f:
        h5f["entry/instrument/detector/data"] = numpy.vstack(
            [frame_1, frame_2, frame_3]
        ).astype(numpy.uint16)

    assert len(compare_volumes(edf_frame_urls, NXtomoScan(scan_2_file, "entry"))) == 1



from __future__ import annotations

from nxtomomill.utils.io import deprecated


class ConfigBase:
    """
    Base class for models that can produce a configuration file (.cfg) to be provided to one of the converter.

    The configuration is setting inputs for the different conversion functions / classes.
    We expect config classes / models to inherit from this class and from a pydantic model.
    """

    def to_cfg_file(self, file_path: str, filter_sections: tuple[str] = tuple()):
        raise NotImplementedError

    @classmethod
    def from_cfg_file(cls, file_path: str) -> ConfigBase:
        raise NotImplementedError

    @deprecated(
        reason="Replaced by pydantic function 'model_dump'", since_version="2.0"
    )
    def to_dict(self) -> dict:
        return self.model_dump()  # pylint: disable=E1101

    @staticmethod
    def from_dict(dict_: dict) -> None:
        """.. warning:: deprecated since 2.0"""
        raise NotImplementedError



"""pint utils"""

from __future__ import annotations

import pint
import logging
from pint.errors import UndefinedUnitError

_logger = logging.getLogger(__name__)

_ureg = pint.get_application_registry()


VALID_CURRENT_VALUES: tuple[str] = [
    str(unit) for unit in (_ureg.ampere, _ureg.kiloampere, _ureg.milliampere)
]
VALID_ENERGY_VALUES: tuple[str] = [
    str(unit)
    for unit in (_ureg.joule, _ureg.eV, _ureg.keV, _ureg.meV, _ureg.GeV, _ureg.kJ)
]
VALID_METRIC_VALUES: tuple[str] = [
    str(unit)
    for unit in (
        _ureg.nanometer,
        _ureg.micrometer,
        _ureg.millimeter,
        _ureg.centimeter,
        _ureg.meter,
    )
]


def get_unit(unit: str, default: pint.Unit, from_dataset: str) -> pint.Unit:
    """
    Convert given unit as an str to a pint unit. Else fallback to 'default'
    """
    if not isinstance(unit, str):
        raise TypeError(f"'unit' is expected to be a {str}. Got {type(unit)}")
    if not isinstance(default, pint.Unit):
        raise TypeError(
            f"'default' is expected to be a {pint.Unit}. Got {type(default)}"
        )

    # special cases (some string that can exists and that are not handled by pint)
    if unit.lower() in ("mev", "megaelectronvolt"):
        unit = _ureg.keV * 1e3
    elif unit.lower() in ("gev", "gigaelectronvolt"):
        unit = _ureg.keV * 1e6
    elif unit == "microns":
        unit = "um"

    try:
        unit = _ureg(unit)
    except UndefinedUnitError:
        _logger.warning(
            f"Undefined unit: {unit} from {from_dataset}. Fallback on default unit {default}"
        )
        return default
    else:
        return unit



import os
import pytest
from glob import glob
from tomoscan.esrf.mock import MockEDF
from tomoscan.esrf.scan.edfscan import EDFTomoScan
from nxtomomill.converter.edf.checks import OUPUT_CHECK
from nxtomomill.models.edf2nx import EDF2nxModel
from nxtomomill.converter.edf.edfconverter import from_edf_to_nx, post_processing_check


def test_edf2nx_check(tmp_path):
    """
    make sure checking an conversion in post processing works
    """

    folder = tmp_path / "test_edf2nx_check"
    folder.mkdir()
    scans = []
    for scan_name, n_proj in zip(["scanA", "scanB"], [10, 11]):
        scan_path = os.path.join(folder, scan_name)
        MockEDF(
            scan_path=scan_path,
            n_radio=n_proj,
            n_ini_radio=n_proj,
            n_extra_radio=0,
            dim=20,
            dark_n=1,
            flat_n=1,
            distance=2.3,
        )
        scans.append(EDFTomoScan(scan_path))

    # do the conversion of the two files
    config_1 = EDF2nxModel()
    config_1.input_folder = scans[0].path
    config_1.output_file = scans[0].path + ".nx"
    from_edf_to_nx(config_1)
    assert os.path.exists(config_1.output_file)

    config_2 = EDF2nxModel()
    config_2.input_folder = scans[1].path
    config_2.output_file = scans[1].path + ".nx"
    from_edf_to_nx(config_2)
    assert os.path.exists(config_2.output_file)

    # check
    post_processing_check(configuration=config_1)
    post_processing_check(configuration=config_2)

    assert os.path.exists(scans[0].path)

    config_test = EDF2nxModel()
    config_test.output_checks = (OUPUT_CHECK.COMPARE_VOLUME,)
    config_test.delete_edf_source_files = True
    config_test.input_folder = scans[0].path
    config_test.output_file = scans[1].path + ".nx"
    assert _get_nb_edf_files(scans[0].path) == 12
    assert _get_nb_edf_files(scans[1].path) == 13

    with pytest.raises(ValueError):
        post_processing_check(configuration=config_test)

    # make sure no edf file have been removed
    assert _get_nb_edf_files(scans[0].path) == 12
    assert _get_nb_edf_files(scans[1].path) == 13

    config_test.input_folder = scans[0].path
    config_test.output_file = scans[0].path + ".nx"

    post_processing_check(configuration=config_test)
    assert _get_nb_edf_files(scans[0].path) == 0
    assert _get_nb_edf_files(scans[1].path) == 13

    config_test.input_folder = scans[1].path
    config_test.output_file = scans[1].path + ".nx"
    post_processing_check(configuration=config_test)
    assert _get_nb_edf_files(scans[1].path) == 0


def _get_nb_edf_files(folder):
    return len(glob(os.path.join(folder, "*.edf")))



from .multitomo import MultiTomoAcquisition
from nxtomomill.utils.io import deprecated_warning


class PCOTomoAcquisition(MultiTomoAcquisition):
    def __init__(self, root_url, configuration, detector_sel_callback, start_index):
        deprecated_warning(
            type_="class",
            name=PCOTomoAcquisition,
            reason="rename PCOTomoAcquisition to MultiTomoAcquisition",
            replacement="nxtomomill.converter.hdf5.acquisition.multitomo.MultiTomoAcquisition",
            since_version="1.2",
        )
        super().__init__(root_url, configuration, detector_sel_callback, start_index)



# coding: utf-8

from __future__ import annotations


from nxtomomill.utils.io import deprecated_warning
from nxtomomill.models.dx2nx import DX2nxModel as _DXFileConfiguration

__all__ = ["DXFileConfiguration"]


class DXFileConfiguration(_DXFileConfiguration):
    def __init__(self, *args, **kwargs):
        deprecated_warning(
            type_="class",
            name="DXFileConfiguration",
            replacement="nxtomomill.models.dx2nx.DX2nxModel",
            reason="moved to models module",
            since_version="2.0",
        )
        super().__init__(*args, **kwargs)



from __future__ import annotations


def group_series(acquisition, list_of_series: list) -> list:
    """
    :param ZSeriesBaseAcquisition acquisition:
    z-series version 2 and 3 are all defined in a separate sequence.
    So we need to aggregate for post processing based on their names.
    post-processing can be dark / flat copy to others NXtomo
    """
    for series in list_of_series:
        if series[0].is_part_of_same_series(acquisition):
            series.append(acquisition)
            return list_of_series
    list_of_series.append(
        [
            acquisition,
        ]
    )
    return list_of_series



"""Application to concatenate a serie of scan with different z to a single NXtomo"""

import argparse
import logging
import os
import re

import numpy as np
from tomoscan.framereducer.method import ReduceMethod

from nxtomomill.converter import from_h5_to_nx
from nxtomomill.io.config import TomoHDF5Config

from ..utils.flat_reducer import flat_reducer
from ..utils.utils import strip_extension

logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__name__)


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def get_arguments(user_args):
    parser = argparse.ArgumentParser(description="""
        creates all the target nxs for all the stages.
        If the postfixes for reference scans are give (the one before and another after the measurements) the reduced flats dark are also created.
        The references scans are expected to contains projections to be 'interpreted' as flats
        """)
    parser.add_argument(
        "--filename_template",
        required=False,
        default=None,
        help="""The filename template. To be used for multiples  zstages it must  contain one or more  segments equal to "X"*ndigits  which will be replaced by the stage number, for the scans, and, for the reference scans, by the begin/end prefixes""",
    )

    parser.add_argument(
        "--filename",
        required=False,
        default=None,
        help="""The filename. To be used for a single scan""",
    )
    parser.add_argument(
        "--output_filename_template",
        required=False,
        default=None,
        help="""Optional, default to a name deduced from filename_template. The output filename template. It must  contain one or more  segments equal to "X"*ndigits  which will be replaced by the stage number""",
    )
    parser.add_argument(
        "--entry_name", required=False, help="entry_name", default="entry0000"
    )

    parser.add_argument(
        "--total_nstages",
        type=int,
        default=None,
        required=True,
        help="How many stages. Example: from 0 to 43 -> --total_nstages  44. ",
    )
    parser.add_argument(
        "--first_stage",
        type=int,
        default=0,
        required=False,
        help="Optional. Defaults to zero. The number of the first considered stage. Use this to do a smaller sequence",
    )
    parser.add_argument(
        "--last_stage",
        type=int,
        default=-1,
        required=False,
        help="Optional. Defaults to total_nstages-1. The number of the last considered stage. Use this to do a smaller sequence",
    )

    parser.add_argument(
        "--do_references",
        type=str2bool,
        default=False,
        required=False,
        help="Optional. If given the reference scans are used for the extraction of the flats/dark. The reference scans are obtained using the ref postfixes",
    )
    parser.add_argument(
        "--extracted_reference_target_dir",
        type=str,
        default=None,
        required=None,
        help="Optional. By default the extracted reference will be written in the same directory as the nexus scan. As the extraction procedure is time consuming they can be written instead to a common directory",
    )
    parser.add_argument(
        "--ref_scan_begin",
        type=str,
        default=None,
        required=False,
        help="""used when "do_reference" is True. It is optional. It is the reference scan. """,
    )

    parser.add_argument(
        "--ref_scan_end",
        type=str,
        default=None,
        required=False,
        help="""used when "do_reference" is True. It is optional. It is the end for the reference scan . """,
    )

    parser.add_argument(
        "--target_directory",
        type=str,
        default="./",
        required=False,
        help="""Where files are written. Optional, defaults to current directory""",
    )
    parser.add_argument(
        "--median_or_mean",
        type=str,
        choices=[ReduceMethod.MEAN.value, ReduceMethod.MEDIAN.value],
        default=ReduceMethod.MEAN.value,
        required=False,
        help="""Choose betwen median or mean. Optional. Default is mean""",
    )
    parser.add_argument(
        "--voxel_size",
        type=float,
        default=None,
        required=False,
        help="""Defaults to zero. If set to a different value the bliss generated nexus files will be corrected. Units are micron""",
    )
    parser.add_argument(
        "--dark_default_value",
        type=float,
        default=300,
        required=False,
        help="""The dark value that is used for scans without dark""",
    )

    args = parser.parse_args(user_args)

    if args.last_stage == -1:
        args.last_stage = args.total_nstages - 1

    return args


def _convert_bliss2nx(bliss_ref_name, nexus_name, corrections_dict={}):
    config = TomoHDF5Config()
    config.overwrite = True
    config.no_master_file = False
    config.input_file = bliss_ref_name
    config.output_file = nexus_name

    for corr_name, corr_value in corrections_dict.items():
        setattr(config, corr_name, corr_value)

    from_h5_to_nx(config)


def main(argv):
    args = get_arguments(argv[1:])

    if args.filename_template is not None:
        name_template_for_numeric = template_to_format_string(args.filename_template)
    else:
        name_template_for_numeric = args.filename

    if name_template_for_numeric is None:
        raise ValueError(
            " Either filename_template of filename must be given as arguments"
        )

    if args.output_filename_template is not None:
        args.output_refname_template = None
        if args.filename is None:
            if args.extracted_reference_target_dir is None:
                args.output_refname_template = template_to_format_string(
                    args.output_filename_template, literal=True
                )
            else:
                args.output_refname_template = None
            args.output_filename_template = template_to_format_string(
                args.output_filename_template, literal=False
            )
    else:
        # will be deduced at output time
        args.output_refname_template = None

    extra_dict = {}
    if args.voxel_size:
        extra_dict.update(
            {
                "x_pixel_size": args.voxel_size * 1.0e-6,
                "y_pixel_size": args.voxel_size * 1.0e-6,
            }
        )

    if args.do_references:
        refs_nexus_names = []
        for bliss_ref_name, what in (
            (args.ref_scan_begin, "begin"),
            (args.ref_scan_end, "end"),
        ):
            if args.output_refname_template is None:
                if args.extracted_reference_target_dir is None:
                    extraction_target_dir = args.target_directory
                else:
                    extraction_target_dir = args.extracted_reference_target_dir

                nexus_name = os.path.join(
                    extraction_target_dir,
                    strip_extension(os.path.basename(bliss_ref_name), _logger) + ".nx",
                )
            else:
                nexus_name = args.output_refname_template.format(what=what)

            _convert_bliss2nx(bliss_ref_name, nexus_name, extra_dict)

            refs_nexus_names.append(nexus_name)

    for iz in range(args.first_stage, args.last_stage + 1):
        bliss_name = name_template_for_numeric.format(i_stage=iz)
        if args.output_filename_template is None:
            nexus_name = os.path.join(
                args.target_directory,
                strip_extension(os.path.basename(bliss_name), _logger) + ".nx",
            )
        else:
            nexus_name = args.output_filename_template.format(i_stage=iz)

        _convert_bliss2nx(bliss_name, nexus_name, extra_dict)

        if args.do_references:
            factor = (iz + 1) / (args.total_nstages)
            flat_reducer(
                nexus_name,
                ref_start_filename=refs_nexus_names[0],
                ref_end_filename=refs_nexus_names[1],
                mixing_factor=factor,
                entry_name=args.entry_name,
                median_or_mean=args.median_or_mean,
                save_intermediated=False,
                reuse_intermediated=True,
                dark_default_value=args.dark_default_value,
            )
    return 0


def template_to_format_string(template, literal=False):
    pattern = re.compile("[X]+")
    # X represent the variable part of the 'template'
    # for example if we want to treat scans HA_2000_sample_0000.nx, ..., HA_2000_sample_9999.nx then
    # we expect the template to be HA_2000_sample_XXXX.nx
    # warning: If the dataset base names contains several X substrings the longest ones will be taken.
    ps = pattern.findall(template)
    ls = list(map(len, ps))
    if len(ls) == 0:
        raise ValueError("The template argument does not contain XX.. segments")
    idx = np.argmax(ls)
    if len(ps[idx]) < 2:
        message = f""" The argument template should contain  a substring  formed by at least two 'X'
        The template was {template}
        """
        raise ValueError(message)
    if not literal:
        template = template.replace(ps[idx], "{i_stage:" + "0" + str(ls[idx]) + "d}")
    else:
        template = template.replace(ps[idx], "{what}")

    return template



# coding: utf-8

import os
import logging
import numpy as np

import h5py

try:
    from nxtomomill.tests.datasets import GitlabDataset
except ImportError:
    raise ImportError("Unable to import GitlabDataset")

from nxtomomill.app.blissfluo2nx import main

logging.disable(logging.INFO)


def test_blissfluo2nx_application_all_dets():
    """test nxtomomill fluo2nx CLI with default parameters (handle all detectors found)"""
    scan_dir = GitlabDataset.get_dataset("blissfluo_datasets3D")
    output_file = os.path.join(scan_dir, "test_Siemens.nx")
    truth_file = os.path.join(scan_dir, "Siemens.nx")
    input_file = os.path.join(scan_dir, "Siemens_raw.h5")

    main(
        [
            "blissfluo2nx",
            input_file,
            output_file,
            "--overwrite",
        ]
    )
    assert os.path.exists(output_file), "output_file doesn't exists"
    with h5py.File(output_file, "r") as f_test, h5py.File(truth_file, "r") as f_true:
        assert len(list(f_test.keys())) == len(
            list(f_true.keys())
        ), f"Number of entries ({len(list(f_test.keys()))}) not as expected ({len(list(f_true.keys()))})."
        assert np.allclose(
            f_test["/grid_Ardesia_ng_mm2_Al_K/data/data"][()],
            f_true["/grid_Ardesia_ng_mm2_Al_K/data/data"][()],
        ), "Discrepancy in grid_Ardesia_Al_K projection data."
        assert np.allclose(
            f_test["/grid_Ardesia_ng_mm2_Al_K/data/rotation_angle"][()],
            f_true["/grid_Ardesia_ng_mm2_Al_K/data/rotation_angle"][()],
        ), "Discrepancy in grid_Ardesia_Al_K rotation angles."


def test_blissfluo2nx_application_single_det():
    """test nxtomomill fluo2nx CLI targetting a single existing detector"""
    scan_dir = GitlabDataset.get_dataset("blissfluo_datasets3D")
    input_file = os.path.join(scan_dir, "Siemens_raw.h5")
    truth_file = os.path.join(scan_dir, "Siemens_1det_2.nx")
    output_file = os.path.join(scan_dir, "test_Siemens_1det.nx")

    main(
        [
            "blissfluo2nx",
            input_file,
            output_file,
            "--detectors",
            "grid_Ardesia_ng_mm2",
            "--overwrite",
        ]
    )
    assert os.path.exists(output_file), "output_file doesn't exists"
    with h5py.File(output_file, "r") as f_test, h5py.File(truth_file, "r") as f_true:
        assert len(list(f_test.keys())) == len(
            list(f_true.keys())
        ), f"Number of entries ({len(list(f_test.keys()))}) not as expected ({len(list(f_true.keys()))})."
        assert np.allclose(
            f_test["/grid_Ardesia_ng_mm2_Al_K/data/data"][()],
            f_true["/grid_Ardesia_ng_mm2_Al_K/data/data"][()],
        ), "Discrepancy in grid_Ardesia_Al_K projection data."
        assert np.allclose(
            f_test["/grid_Ardesia_ng_mm2_Al_K/data/rotation_angle"][()],
            f_true["/grid_Ardesia_ng_mm2_Al_K/data/rotation_angle"][()],
        ), "Discrepancy in grid_Ardesia_Al_K rotation angles."


def test_blissfluo2nx_application_two_det():
    """test nxtomomill fluo2nx CLI targetting two existing detector"""
    scan_dir = GitlabDataset.get_dataset("blissfluo_datasets3D")
    input_file = os.path.join(scan_dir, "Siemens_raw.h5")
    truth_file = os.path.join(scan_dir, "Siemens_2det_2.nx")
    output_file = os.path.join(scan_dir, "test_Siemens_2det.nx")

    main(
        [
            "blissfluo2nx",
            input_file,
            output_file,
            "--detectors",
            "grid_Ardesia_ng_mm2",
            "grid_weighted_ng_mm2",
            "--overwrite",
        ]
    )
    assert os.path.exists(output_file), "output_file doesn't exists"
    with h5py.File(output_file, "r") as f_test, h5py.File(truth_file, "r") as f_true:
        assert len(list(f_test.keys())) == len(
            list(f_true.keys())
        ), f"Number of entries ({len(list(f_test.keys()))}) not as expected ({len(list(f_true.keys()))})."
        assert np.allclose(
            f_test["/grid_Ardesia_ng_mm2_Al_K/data/data"][()],
            f_true["/grid_Ardesia_ng_mm2_Al_K/data/data"][()],
        ), "Discrepancy in grid_Ardesia_Al_K projection data."
        assert np.allclose(
            f_test["/grid_Ardesia_ng_mm2_Al_K/data/rotation_angle"][()],
            f_true["/grid_Ardesia_ng_mm2_Al_K/data/rotation_angle"][()],
        ), "Discrepancy in grid_Ardesia_Al_K rotation angles."



from __future__ import annotations

from pydantic import BaseModel, Field, field_validator, ConfigDict
from nxtomomill.settings import Tomo
from nxtomomill.models.utils import convert_str_to_tuple, is_url_path, filter_str_def


from silx.io.url import DataUrl


class EntriesAndTitlesSection(BaseModel):

    model_config = ConfigDict(
        validate_assignment=True, arbitrary_types_allowed=True, validate_by_name=True
    )

    entries: tuple[DataUrl, ...] = Field(
        default=tuple(), description="List of root/init entries to convert"
    )
    sub_entries_to_ignore: tuple[DataUrl, ...] = Field(
        default=tuple(), description="List of sub entries to ignore"
    )
    init_titles: tuple[str, ...] = Field(
        default=Tomo.H5.INIT_TITLES, description="Titles for initialization"
    )
    zseries_init_titles: tuple[str, ...] = Field(
        default=Tomo.H5.ZSERIE_INIT_TITLES,
        description="Titles for z-serie initialization",
        alias="zserie_init_titles",
    )
    multitomo_init_titles: tuple[str, ...] = Field(
        default=Tomo.H5.MULTITOMO_INIT_TITLES,
        description="Titles for multi-tomo initialization",
    )
    back_and_forth_init_titles: tuple[str, ...] = Field(
        default=Tomo.H5.BACK_AND_FORTH_INIT_TITLES,
        description="Titles for back-and-forth initialization",
    )
    dark_titles: tuple[str, ...] = Field(
        default=Tomo.H5.DARK_TITLES,
        description="Titles to determine if the bliss scan is a set of dark",
    )
    flat_titles: tuple[str, ...] = Field(
        default=Tomo.H5.FLAT_TITLES,
        description="Titles to determine if the bliss scan is a set of flat",
    )
    projection_titles: tuple[str, ...] = Field(
        default=Tomo.H5.PROJ_TITLES,
        description="Titles to determine if the bliss scan is a set of projection",
        alias="proj_titles",
    )
    alignment_titles: tuple[str, ...] = Field(
        default=Tomo.H5.ALIGNMENT_TITLES,
        description="Titles to determine if the bliss scan is a set of alignment projection",
    )

    @field_validator(
        "init_titles",
        "zseries_init_titles",
        "multitomo_init_titles",
        "back_and_forth_init_titles",
        "dark_titles",
        "flat_titles",
        "projection_titles",
        "alignment_titles",
        mode="plain",
    )
    @classmethod
    def cast_entries_or_title_to_tuple(cls, value: str | tuple[str]) -> tuple[str, ...]:
        if isinstance(value, str):
            value = filter_str_def(value)
        return convert_str_to_tuple(value) or ()

    @field_validator(
        "entries",
        "sub_entries_to_ignore",
        mode="before",
    )
    @classmethod
    def cast_entries_or_title_to_tuple_of_url(
        cls, value: str | tuple[DataUrl | str, ...]
    ) -> tuple[DataUrl, ...]:
        """The entries are first converted from a string to a list of string.
        Then this list of string will be processed later to create a list of DataUrl with relative link and with a context
        """
        if isinstance(value, str):
            value = filter_str_def(value)
            value = convert_str_to_tuple(value)
        entries = EntriesAndTitlesSection.parse_frame_urls(value)
        return tuple(
            [EntriesAndTitlesSection.fix_entry_name(entry) for entry in entries]
        )

    @staticmethod
    def parse_frame_urls(urls: tuple[str | DataUrl, ...]):
        """
        Insure urls is None or a list of valid DataUrl
        """
        if urls in ("", None):
            return tuple()
        res = []
        for i_url, url in enumerate(urls):
            if isinstance(url, str):
                if url == "":
                    continue
                elif is_url_path(url):
                    url = DataUrl(path=url)
                else:
                    url = DataUrl(data_path=url, scheme="silx")
            if not isinstance(url, DataUrl):
                raise ValueError(
                    "urls tuple should contains DataUrl. "
                    f"Not {type(url)} at index {i_url}"
                )
            else:
                res.append(url)
        return tuple(res)

    @staticmethod
    def fix_entry_name(entry: DataUrl):
        """simple util function to insure the entry start by a "/"""
        if not isinstance(entry, DataUrl):
            raise TypeError("entry is expected to be a DataUrl")
        if not entry.data_path().startswith("/"):
            entry = DataUrl(
                scheme=entry.scheme(),
                data_slice=entry.data_slice(),
                file_path=entry.file_path(),
                data_path="/" + entry.data_path(),
            )
        return entry



# coding: utf-8

import os
import logging

import h5py

from tomoscan.tests.datasets import GitlabDataset
from nxtomomill.app.fluo2nx import main

logging.disable(logging.INFO)


def test_fluo2nx_application_all_dets(tmp_path):
    """test nxtomomill fluo2nx CLI with default parameters (handle all detectors found)"""
    scan_dir = GitlabDataset.get_dataset("fluo_datasets")
    output_file = os.path.join(tmp_path, "nexus_file.nx")

    assert not os.path.exists(output_file), "output_file exists already"
    main(["fluo2nx", scan_dir, output_file, "CP1_XRD_insitu_top_ft_100nm"])
    assert os.path.exists(output_file), "output_file doesn't exists"
    with h5py.File(output_file, "r") as f:
        assert (
            len(list(f.keys())) == 45
        ), f"Number of entries ({len(list(f.keys()))}) not as expected (45)."


def test_fluo2nx_application_single_det(tmp_path):
    """test nxtomomill fluo2nx CLI targetting a single existing detector"""
    scan_dir = GitlabDataset.get_dataset("fluo_datasets")
    output_file = os.path.join(tmp_path, "nexus_file.nx")

    assert not os.path.exists(output_file), "output_file exists already"
    main(
        [
            "fluo2nx",
            scan_dir,
            output_file,
            "CP1_XRD_insitu_top_ft_100nm",
            "--detectors",
            "xmap",
        ]
    )
    assert os.path.exists(output_file), "output_file doesn't exists"
    with h5py.File(output_file, "r") as f:
        assert (
            len(list(f.keys())) == 15
        ), f"Number of entries ({len(list(f.keys()))}) not as expected (15)."


def test_fluo2nx_application_two_dets(tmp_path):
    """test nxtomomill fluo2nx CLI targetting two existing detectors"""
    scan_dir = GitlabDataset.get_dataset("fluo_datasets")
    output_file = os.path.join(tmp_path, "nexus_file.nx")

    assert not os.path.exists(output_file), "output_file exists already"
    main(
        [
            "fluo2nx",
            scan_dir,
            output_file,
            "CP1_XRD_insitu_top_ft_100nm",
            "--detectors",
            "xmap",
            "falcon",
        ]
    )
    assert os.path.exists(output_file), "output_file doesn't exists"
    with h5py.File(output_file, "r") as f:
        assert (
            len(list(f.keys())) == 30
        ), f"Number of entries ({len(list(f.keys()))}) not as expected (30)."


def test_fluo2nx_application_2D(tmp_path):
    scan_dir = GitlabDataset.get_dataset("fluo_datasets2D")
    output_file = os.path.join(tmp_path, "nexus_file_2d.nx")

    assert not os.path.exists(output_file), "output_file exists already"
    main(
        [
            "fluo2nx",
            scan_dir,
            output_file,
            "CONT2_p2_600nm_FT02_slice_0",
            "--dimension",
            "2",
        ]
    )
    assert os.path.exists(output_file), "output_file doesn't exists"
    with h5py.File(output_file, "r") as f:
        assert (
            len(list(f.keys())) == 4
        ), f"Number of entries ({len(list(f.keys()))}) not as expected (4)."


def test_fluo2nx_application_2D_single_det(tmp_path):
    """test nxtomomill fluo2nx CLI targetting a single existing detector"""
    scan_dir = GitlabDataset.get_dataset("fluo_datasets2D")
    output_file = os.path.join(tmp_path, "nexus_file_2D_singledet.nx")

    assert not os.path.exists(output_file), "output_file exists already"
    main(
        [
            "fluo2nx",
            scan_dir,
            output_file,
            "CONT2_p2_600nm_FT02_slice_0",
            "--detectors",
            "corrweighted",
            "--dimension",
            "2",
        ]
    )
    assert os.path.exists(output_file), "output_file doesn't exists"
    with h5py.File(output_file, "r") as f:
        assert (
            len(list(f.keys())) == 2
        ), f"Number of entries ({len(list(f.keys()))}) not as expected (2)."



"""
Application to check the conversion done of an EDF folder to a .nx file.

.. program-output:: nxtomomill edf2nx-check --help

"""

import argparse
import logging

from nxtomomill.converter.edf.edfconverter import post_processing_check
from nxtomomill.models.utils import convert_str_to_tuple
from nxtomomill.io.config.edfconfig import TomoEDFConfig
from nxtomomill.converter.edf.checks import OUPUT_CHECK

logging.basicConfig(level=logging.INFO)


def main(argv):
    """
    check a conversion made.
    To keep it coherent and safe we must keep the same inputs as for the EDF to NXtomo conversion.
    The only difference is that this one has some output check defined by default
    """
    parser = argparse.ArgumentParser(
        description="check a conversion from edf to NXtomo went's well"
    )
    parser.add_argument("scan_path", help="folder containing the edf files", nargs="?")
    parser.add_argument("output_file", help="foutput .h5 file", nargs="?")
    parser.add_argument(
        "--dataset-basename",
        "--file-prefix",
        default=None,
        help="file prefix to be used to deduce projections",
    )
    parser.add_argument(
        "--info-file",
        default=None,
        help=".info file containing acquisition information (ScanRange, Energy, TOMO_N...)",
    )
    parser.add_argument(
        "--config",
        "--configuration-file",
        "--configuration",
        default=None,
        help="file containing the full configuration to convert from SPEC-EDF to bliss to nexus. "
        "Default configuration file can be created from `nxtomomill edf-config` command",
    )
    parser.add_argument(
        "--delete-edf",
        "--delete-edf-source",
        "--delete-edf-source-files",
        default=False,
        action="store_true",
        help="if you are duplicating data (default behavior) you can ask to delete used edf files to free space",
    )
    parser.add_argument(
        "--output-checks",
        default=(OUPUT_CHECK.COMPARE_VOLUME,),
        help="Define list of check to be run once the conversion is finished (raise an error if check fails). This is done before removing edf if asked. So if check fails source edf files won't be removed",
    )
    options = parser.parse_args(argv[1:])
    config = TomoEDFConfig()
    if options.config is not None:
        config = config.from_cfg_file(options.config)

    check_input = {
        "dataset basename": (options.dataset_basename, config.dataset_basename),
        "scan path": (options.scan_path, config.input_folder),
        "output file": (options.output_file, config.output_file),
        "info file": (options.info_file, config.dataset_info_file),
    }
    for input_name, (opt_value, config_value) in check_input.items():
        if (
            opt_value is not None
            and config_value is not None
            and opt_value != config_value
        ):
            raise ValueError(
                f"two values for {input_name} are provided from arguments and configuration file ({opt_value, config_value})"
            )

    if options.dataset_basename is not None:
        config.dataset_basename = options.dataset_basename
    if options.info_file is not None:
        config.dataset_info_file = options.info_file
    if options.scan_path is not None:
        config.input_folder = options.scan_path
    if options.output_file is not None:
        config.output_file = options.output_file
    config.delete_edf_source_files = options.delete_edf
    config.output_checks = convert_str_to_tuple(options.output_checks)
    if not len(config.output_checks) > 0:
        raise ValueError("not check on conversion selected")

    assert isinstance(
        config.output_checks, tuple
    ), f"config.output_checks is expected to be a tuple. Gets {type(config.output_checks)}"
    post_processing_check(
        configuration=config,
    )



# coding: utf-8

from datetime import datetime

import numpy.random
from tomoscan.io import HDF5File


class MockDxFile:
    """
    Mock DXFile
    """

    def __init__(
        self,
        file_path,
        n_projection,
        n_darks,
        n_flats,
        det_height=128,
        det_width=128,
        data_path="/",
    ):
        self._file_path = file_path
        self._n_projection = n_projection
        self._n_darks = n_darks
        self._n_flats = n_flats
        self._det_height = det_height
        self._det_width = det_width
        self._n_projection = n_projection
        self.data_flat = None
        self.data_proj = None
        self.data_dark = None

        with HDF5File(file_path, mode="w") as h5f:
            root_grp = h5f.require_group(data_path)
            exchange_grp = root_grp.require_group("exchange")
            # create data
            self.data_proj = self._create_random_frames(self._n_projection)
            exchange_grp["data"] = self.data_proj
            # create data_dark
            if self._n_darks > 0:
                self.data_dark = self._create_random_frames(self._n_darks)
                exchange_grp["data_dark"] = self.data_dark
            # create data_white
            if self._n_flats > 0:
                self.data_flat = self._create_random_frames(self._n_flats)
                exchange_grp["data_white"] = self.data_flat
            root_grp["file_creation_datetime"] = str(datetime.now())
            # for now some information are not mock and used like
            # exposure_period, x_binning, roi... did not had a concrete
            # example about it

    def _create_random_frames(self, n_frames):
        data = numpy.random.random(self._det_height * self._det_width * n_frames) * 256
        data = data.astype(numpy.uint16)
        return data.reshape((n_frames, self._det_height, self._det_width))



# coding: utf-8

from __future__ import annotations

from configparser import ConfigParser

from pydantic import BaseModel, ConfigDict

from ..base.FrmFlatToNestedBase import FrmFlatToNestedBase
from ..base.NestedModelBase import NestedModelBase

from ..base.source_section import SourceSection as _SourceSection
from ..base.instrument_section import InstrumentSection
from .general_section import GeneralSection
from nxtomomill.utils.io import deprecated_warning

__all__ = ["Fluo2nxModel", "generate_default_fluo_config"]


class _LegacySourceSection(_SourceSection, InstrumentSection):
    """Historically the instrument name and section was melt with the source section"""

    pass


class Fluo2nxModel(FrmFlatToNestedBase, GeneralSection, _LegacySourceSection):
    """
    Configuration class to provide to the fluo->nx converter .
    """

    model_config = ConfigDict(validate_assignment=True)

    def to_nested_model(self) -> _NestedTomoFluoConfig:
        return _NestedTomoFluoConfig(
            general_section=GeneralSection(
                output_file=self.output_file,
                dataset_basename=self.dataset_basename,
                dataset_info_file=self.dataset_info_file,
                detector_names=self.detector_names,
                dimension=self.dimension,
                duplicate_data=self.duplicate_data,
                file_extension=self.file_extension,
                input_folder=self.input_folder,
                log_level=self.log_level,
                overwrite=self.overwrite,
                patterns_to_ignores=self.patterns_to_ignores,
            ),
            source_section=_LegacySourceSection(
                instrument_name=self.instrument_name,
                source_name=self.source_name,
                source_probe=self.source_probe,
                source_type=self.source_type,
            ),
        )

    @classmethod
    def from_cfg_file(cls, file_path: str) -> Fluo2nxModel:
        txt_parser = ConfigParser(allow_no_value=True)
        txt_parser.read(file_path)

        def get_section(name, default={}):
            if txt_parser.has_section(name):
                return txt_parser[name]
            else:
                return default

        return _NestedTomoFluoConfig(
            general_section=GeneralSection(**get_section("GENERAL_SECTION")),
            source_section=_LegacySourceSection(**get_section("SOURCE_SECTION")),
        ).to_flatten_config()

    @staticmethod
    def from_dict(dict_: dict) -> None:
        deprecated_warning(
            type_="function",
            name="from_dict",
            reason="replaced by pydantic 'model_dump' function",
            replacement="model_dump",
            since_version="2.0",
        )

        dict_ = {key.lower(): value for key, value in dict_.items()}
        config = _NestedTomoFluoConfig(**dict_)
        return config.to_flatten_config()


class _NestedTomoFluoConfig(BaseModel, NestedModelBase):

    model_config = ConfigDict(str_to_upper=True)

    general_section: GeneralSection = GeneralSection()
    source_section: _LegacySourceSection = _LegacySourceSection()

    def to_flatten_config(self) -> Fluo2nxModel:
        return Fluo2nxModel(
            **self.general_section.model_dump(),
            **self.source_section.model_dump(),
        )


def generate_default_fluo_config(level: str = "required") -> dict:
    """generate a default configuration to convert fluo-tomo data (after PyMCA fit) to NXtomo"""
    config = Fluo2nxModel().to_nested_model()
    return {key.upper(): value for key, value in config.model_dump().items()}



# coding: utf-8

import contextlib

import h5py
import pint
import logging
from .pintutils import get_unit

try:
    import hdf5plugin  # noqa F401
except ImportError:
    pass
from silx.io.url import DataUrl
from silx.io.utils import open as open_hdf5

_logger = logging.getLogger(__name__)

__all__ = ["EntryReader", "DatasetReader"]


class _BaseReader(contextlib.AbstractContextManager):
    def __init__(self, url: DataUrl):
        if not isinstance(url, DataUrl):
            raise TypeError(f"url should be an instance of DataUrl. Not {type(url)}")
        if url.scheme() not in ("silx", "h5py"):
            raise ValueError("Valid scheme are silx and h5py")
        if url.data_slice() is not None:
            raise ValueError(
                "Data slices are not managed. Data path should "
                "point to a bliss node (h5py.Group)"
            )
        self._url = url
        self._file_handler = None

    def __exit__(self, *exc):
        return self._file_handler.close()


class EntryReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = open_hdf5(filename=self._url.file_path())
        if self._url.data_path() == "":
            entry = self._file_handler
        elif self._url.data_path() not in self._file_handler:
            raise KeyError(
                f"data path '{self._url.data_path()}' doesn't exists from '{self._url.file_path()}'"
            )
        else:
            entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Group):
            raise ValueError("Data path should point to a bliss node (h5py.Group)")
        return entry


class DatasetReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = open_hdf5(filename=self._url.file_path())
        entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Dataset):
            raise ValueError(
                f"Data path ({self._url.path()}) should point to a dataset (h5py.Dataset)"
            )
        return entry


def get_dataset_unit(
    dataset: h5py.Dataset, default: pint.Unit, from_dataset: str
) -> pint.Unit:
    """
    Util function to return the pint Unit of a HDF5 dataset.
    This dataset must have `unit` or `units` defined else will fall back to the default unit

    :param from_dataset: information about the 'dataset' / metadata we are trying to access. For logging purpose in case of failure.
    """
    if not isinstance(dataset, h5py.Dataset):
        raise TypeError(
            f"dataset is expected to be an instance of {h5py.Dataset}. Got {type(dataset)}."
        )
    if "unit" in dataset.attrs:
        unit = dataset.attrs["unit"]
    elif "units" in dataset.attrs:
        unit = dataset.attrs["units"]
    else:
        _logger.info(f"no unit found for {from_dataset}. Take default unit: {default}")
        return default

    if hasattr(unit, "decode"):
        # handle Diamond dataset
        unit = unit.decode()

    return get_unit(unit=unit, default=default, from_dataset=from_dataset)



from nxtomomill.utils.io import deprecated_warning
from nxtomo.nxobject.nxtransformations import *  # noqa F401

deprecated_warning(
    type_="Module",
    name="nxtomomill.nexus.nxtransformations",
    reason="dedicated project created",
    replacement="nxtomo.nxobject.nxtransformations",
    since_version=1.0,
)



# coding: utf-8

"""
module to convert from dx file (hdf5) to nexus tomo compliant .nx (hdf5)
"""

from __future__ import annotations

import logging
import os
import pint

import h5py
import numpy
from silx.io.url import DataUrl
from silx.io.utils import get_data, h5py_read_dataset
from tomoscan.io import HDF5File
from nxtomo.nxobject.nxdetector import FieldOfView, ImageKey
from nxtomo.utils.frameappender import FrameAppender

from nxtomomill.converter.baseconverter import BaseConverter
from nxtomomill.converter.version import version as converter_version
from nxtomomill.models.dx2nx import DX2nxModel

from nxtomomill.utils.hdf5 import DatasetReader, EntryReader, get_dataset_unit
from silx.utils.deprecation import deprecated

_ureg = pint.get_application_registry()
_logger = logging.getLogger(__name__)


__all__ = ["from_dx_to_nx", "from_dx_config_to_nx"]


# dev note: to ensure backward compatibility we needed to keep from_dx_to_nx. And we created from_dx_config_to_nx to handle the configuration file.
# but with time we expect to remove `from_dx_to_nx` and them rename from_dx_config_to_nx to `from_dx_to_nx` in order to get something coherent.
@deprecated(
    replacement="from_dx_config_to_nx",
    since_version="0.13",
    reason="Some configuration dedicated classes exists for tuning the configuration directly instead of provided n parameter to the function.",
)
def from_dx_to_nx(
    input_file: str,
    output_file: str | None = None,
    file_extension: str = ".nx",
    duplicate_data: bool = True,
    input_entry: str = "/",
    output_entry: str = "entry0000",
    scan_range: tuple = (0.0, 180.0),
    pixel_size=(None, None),
    field_of_view: FieldOfView | str | None = None,
    distance: float | None = None,
    overwrite: bool = True,
    energy: float | None = None,
) -> tuple:
    """

    :param input_file: dxfile to convert
    :param output_file: output file to save
    :param file_extension: file extension to give if the output file is
        not provided.
    :param duplicate_data: if True frames will be duplicated. Otherwise
        we will create (relative) link to the input file.
    :param input_entry: path to the HDF5 group to convert. For now it looks
        each file can only contain one dataset. Just to ensure any future
        compatibility if it evolve with time.
    :param output_entry: path to store the NxTomo created.
    :param scan_range: tuple of two elements with the minimum scan range.
        Projections are expected to be taken with equal angular spaces.
    :param pixel_size: pixel size can be provided (in meter and as
        x_pizel_size, y_pixel_size)
    :param field_of_view: field of view
    :param distance: sample / detector distance in meter
    :param overwrite: if True and if the entry already exists in the
        output file then will overwrite it.
    :return: tuple of (output_file, entry) created. For now the list should
        contain at most one of those tuple
    """
    configuration = DX2nxModel(input_file=input_file, output_file=output_file)
    configuration.file_extension = file_extension
    configuration.copy_data = duplicate_data
    configuration.input_entry = input_entry
    configuration.output_entry = output_entry
    configuration.scan_range = scan_range
    configuration.pixel_size = pixel_size
    configuration.field_of_view = field_of_view
    configuration.sample_detector_distance = distance
    configuration.overwrite = overwrite
    configuration.energy = energy
    return from_dx_config_to_nx(configuration=configuration)


def from_dx_config_to_nx(
    configuration: DX2nxModel,
):
    """
    Convert from dxfile to NXtomo.
    Dark and flats will be store at the beginning and we consider they are
    take at start so rotation angle will set to scan_range[0].
    Projection rotation angle will be interpolated from scan_range and
    with equality space distribution.
    We do not expect any alignment projection.
    """
    converter = _DxFileToNxConverter(configuration=configuration)
    return converter.convert()


class _PathDoesNotExistsInExchange(Exception):
    pass


class _DxFileToNxConverter(BaseConverter):
    """
    Convert from dxfile to NXtomo.
    Dark and flats will be store at the beginning and we consider they are
    take at start so rotation angle will set to scan_range[0].
    Projection rotation angle will be interpolated from scan_range and
    with equality space distribution.
    We do not expect any alignment projection.
    """

    DEFAULT_SAMPLE_DETECTOR_DISTANCE_VALUE: float = 1.0

    DEFAULT_PIXEL_VALUE: float = 1.0

    DEFAULT_BEAM_ENERGY: float = 1.0

    def __init__(
        self,
        configuration: DX2nxModel,
    ):
        self._configuration = configuration
        if not len(self.scan_range) == 2:
            raise ValueError("scan_range expects to be a tuple with two elements")

        input_file = os.path.abspath(self.input_file)
        if self.output_file is None:
            input_file_basename, _ = os.path.splitext(input_file)
            if not self._configuration.file_extension.startswith("."):
                self._configuration.file_extension = (
                    "." + self._configuration.file_extension
                )
            output_file = os.path.join(
                os.path.dirname(input_file),
                os.path.basename(input_file_basename)
                + self._configuration.file_extension,
            )
        else:
            output_file = os.path.abspath(self.output_file)
        self._configuration.output_file = output_file
        if self.input_entry == "/":
            self._configuration._input_entry = ""
        else:
            self._configuration._input_entry = self.input_entry
        if self.field_of_view is not None:
            fov = self.field_of_view
            if isinstance(fov, str):
                fov = fov.title()
            self._configuration.field_of_view = FieldOfView(fov)

        self._n_frames = 0
        self._data_proj_url = None
        self._data_darks_url = None
        self._data_flats_url = None
        self._input_root_url = DataUrl(
            file_path=self.input_file,
            data_path=self.input_entry,
            scheme="silx",
        )

    @property
    def input_file(self):
        return self._configuration.input_file

    @property
    def input_entry(self):
        return self._configuration.input_entry

    @property
    def output_file(self):
        return self._configuration.output_file

    @property
    def output_entry(self):
        return self._configuration.output_entry

    @property
    def scan_range(self):
        return self._configuration.scan_range

    @property
    def copy_data(self):
        return self._configuration.copy_data

    @property
    def overwrite(self):
        return self._configuration.overwrite

    @property
    def field_of_view(self) -> FieldOfView | None:
        return self._configuration.field_of_view

    @property
    def energy(self) -> float | None:
        return self._configuration.energy

    @property
    def input_root_url(self):
        return self._input_root_url

    def convert(self):
        """
        do conversion from dxfile to NXtomo

        :return: tuple of (output_file, entry) created. For now the list should
                 contain at most one of those tuple
        """
        with HDF5File(self.output_file, mode="a") as h5f:
            if self.output_entry in h5f:
                if self.overwrite:
                    del h5f[self.output_entry]
                else:
                    raise OSError(
                        "{} already exists cannot create requested NXtomo entry. Won't overwrite it as not requested"
                    )

        self._n_frames = 0
        self._data_proj_url = DataUrl(
            file_path=self.input_file,
            data_path="/".join((self.input_entry, "exchange", "data")),
            scheme="silx",
        )
        self._data_darks_url = DataUrl(
            file_path=self.input_file,
            data_path="/".join((self.input_entry, "exchange", "data_dark")),
            scheme="silx",
        )
        self._data_flats_url = DataUrl(
            file_path=self.input_file,
            data_path="/".join((self.input_entry, "exchange", "data_white")),
            scheme="silx",
        )

        # convert frames
        if self.copy_data:
            self._convert_frames_with_duplication()
        else:
            self._convert_frames_without_duplication()

        # convert detector extra information
        #  x pixel size
        x_pixel_size, y_pixel_size = self._configuration.pixel_size
        if x_pixel_size is None:
            try:
                x_pixel_size = self._read_x_pixel_size().to_base_units().magnitude
            except Exception:
                x_pixel_size = self.DEFAULT_PIXEL_VALUE
                _logger.warning(
                    "No x pixel size found or provided. Set the it to the default value"
                )
        with HDF5File(self.output_file, mode="a") as h5f:
            root_grp = h5f.require_group(self.output_entry)
            instrument_grp = root_grp.require_group("instrument")
            detector_grp = instrument_grp.require_group("detector")
            detector_grp["x_pixel_size"] = x_pixel_size
            detector_grp["x_pixel_size"].attrs["units"] = "m"
        #  y pixel size
        if y_pixel_size is None:
            try:
                y_pixel_size = self._read_y_pixel_size().to_base_units().magnitude
            except Exception:
                y_pixel_size = self.DEFAULT_PIXEL_VALUE
                _logger.warning(
                    "No y pixel size found or provided. Set the it to the default value"
                )

        with HDF5File(self.output_file, mode="a") as h5f:
            root_grp = h5f.require_group(self.output_entry)
            instrument_grp = root_grp.require_group("instrument")
            detector_grp = instrument_grp.require_group("detector")
            detector_grp["y_pixel_size"] = y_pixel_size
            detector_grp["y_pixel_size"].attrs["units"] = "m"
        #  field of view
        if self.field_of_view is not None:
            with HDF5File(self.output_file, mode="a") as h5f:
                root_grp = h5f.require_group(self.output_entry)
                instrument_grp = root_grp.require_group("instrument")
                detector_grp = instrument_grp.require_group("detector")
                detector_grp["field_of_view"] = self.field_of_view.value
        #  distance
        if self._configuration.sample_detector_distance is None:
            # if value not set by the user let read it
            try:
                self._configuration.sample_detector_distance = (
                    self._read_sample_detector_distance().to_base_units().magnitude
                )
            except Exception:
                self._configuration.sample_detector_distance = (
                    self.DEFAULT_SAMPLE_DETECTOR_DISTANCE_VALUE
                )
                _logger.warning(
                    "No detector / sample distance found or provided. "
                    "Set the it  the default value"
                )

        with HDF5File(self.output_file, mode="a") as h5f:
            root_grp = h5f.require_group(self.output_entry)
            instrument_grp = root_grp.require_group("instrument")
            detector_grp = instrument_grp.require_group("detector")
            detector_grp["distance"] = self._configuration.sample_detector_distance
            detector_grp["distance"].attrs["units"] = str(_ureg.meter)
        # energy
        if self.energy is None:
            try:
                self._configuration.energy = self._read_energy().to(_ureg.keV).magnitude
            except Exception:
                self._configuration.energy = self.DEFAULT_BEAM_ENERGY
                _logger.warning(
                    "No energy found or provided. " "Set the it to the default value"
                )

        with HDF5File(self.output_file, mode="a") as h5f:
            root_grp = h5f.require_group(self.output_entry)
            beam_grp = root_grp.require_group("beam")
            beam_grp["incident_energy"] = self._configuration.energy
            beam_grp["incident_energy"].attrs["units"] = str(_ureg.keV)

        #  count_time
        self._copy_count_time()

        # start time
        with EntryReader(self.input_root_url) as input_entry:
            if "file_creation_datetime" in input_entry:
                with HDF5File(self.output_file, mode="a") as h5f:
                    root_grp = h5f.require_group(self.output_entry)
                    root_grp["start_time"] = h5py_read_dataset(
                        input_entry["file_creation_datetime"]
                    )

        return ((self.output_file, self.output_entry),)

    def _copy_count_time(self):
        """Try to read and copy count_time / exposure_period"""

        def read_and_write_count_time(dataset: h5py.Dataset):
            count_time = h5py_read_dataset(dataset)
            if count_time == "Error: Unknown Attribute":
                _logger.info(
                    f"Count time not stored on {self.input_entry}@{self.input_file}"
                )
            else:
                try:
                    if len(count_time) == self._n_frames:
                        with HDF5File(self.output_file, mode="a") as h5f:
                            root_grp = h5f.require_group(self.input_entry)
                            instrument_grp = root_grp.require_group("instrument")
                            detector_grp = instrument_grp.require_group("detector")
                            detector_grp["count_time"] = count_time
                    else:
                        _logger.warning(
                            f"exposure period and data frame have an incoherent size ({len(count_time)} vs {self._n_frames})"
                        )
                except Exception as e:
                    _logger.error(
                        f"Failed to get 'count_time' / 'exposure_period'. reason is {e}"
                    )

        with EntryReader(self.input_root_url) as input_entry:
            if "measurement" in input_entry:
                measurement_grp = input_entry["measurement"]
                if "detector" in measurement_grp:
                    detector_grp = measurement_grp["detector"]
                    if "exposure_period" in detector_grp:
                        read_and_write_count_time(detector_grp["exposure_period"])

    def _convert_frames_with_duplication(self):
        image_key = []
        image_key_control = []
        rotation_angle = []
        data = None

        # handle darks
        try:
            data_dark = get_data(self._data_darks_url)
        except _PathDoesNotExistsInExchange:
            _logger.warning(f"No darks found in {self.input_entry}@{self.input_file}")
        else:
            image_key.extend([ImageKey.DARK_FIELD.value] * len(data_dark))
            image_key_control.extend([ImageKey.DARK_FIELD.value] * len(data_dark))
            rotation_angle.extend([self.scan_range[0]] * len(data_dark))
            if data is None:
                data = data_dark
            else:
                data = numpy.concatenate((data, data_dark), axis=0)
        # handle flats
        try:
            data_flat = get_data(self._data_flats_url)
        except _PathDoesNotExistsInExchange:
            _logger.warning(f"No flats found in {self.input_entry}@{self.input_file}")
        else:
            image_key.extend([ImageKey.FLAT_FIELD.value] * len(data_flat))
            image_key_control.extend([ImageKey.FLAT_FIELD.value] * len(data_flat))
            rotation_angle.extend([self.scan_range[0]] * len(data_flat))
            if data is None:
                data = data_flat
            else:
                data = numpy.concatenate((data, data_flat), axis=0)
        # handle projections
        data_proj = get_data(self._data_proj_url)
        image_key.extend([ImageKey.PROJECTION.value] * len(data_proj))
        image_key_control.extend([ImageKey.PROJECTION.value] * len(data_proj))
        assert (
            self.scan_range[0] is not None
        ), "scan range is expected to be a tuple of float"
        assert (
            self.scan_range[1] is not None
        ), "scan range is expected to be a tuple of float"
        rotation_angle.extend(
            numpy.linspace(
                self.scan_range[0],
                self.scan_range[1],
                num=len(data_proj),
                endpoint=True,
            )
        )
        if data is None:
            data = data_proj
        else:
            data = numpy.concatenate((data, data_proj), axis=0)
        self._n_frames = len(data)

        with HDF5File(self.output_file, mode="a") as h5f:
            root_grp = h5f.require_group(self.output_entry)
            instrument_grp = root_grp.require_group("instrument")
            detector_grp = instrument_grp.require_group("detector")
            detector_grp["data"] = data
            detector_grp["image_key"] = image_key
            detector_grp["image_key_control"] = image_key_control
            sample_grp = root_grp.require_group("sample")
            sample_grp["rotation_angle"] = rotation_angle
            sample_grp["rotation_angle"].attrs["units"] = str(_ureg.degree)
            self._path_nxtomo_attrs(root_grp)

    def _path_nxtomo_attrs(self, root_grp):
        root_grp.attrs["NX_class"] = "NXentry"
        root_grp.attrs["definition"] = "NXtomo"
        root_grp.attrs["version"] = converter_version()
        root_grp.attrs["default"] = "instrument/detector"
        instrument_grp = root_grp.require_group("instrument")
        instrument_grp.attrs["NX_class"] = "NXinstrument"
        instrument_grp.attrs["default"] = "detector"
        detector_grp = instrument_grp.require_group("detector")
        detector_grp.attrs["NX_class"] = "NXdetector"
        detector_grp.attrs["NX_class"] = "NXdata"
        detector_grp.attrs["signal"] = "data"
        detector_grp.attrs["SILX_style/axis_scale_types"] = ["linear", "linear"]
        if "data" in detector_grp:
            detector_grp["data"].attrs["interpretation"] = "image"
        sample_node = root_grp.require_group("sample")
        sample_node.attrs["NX_class"] = "NXsample"

    def _convert_frames_without_duplication(self):
        image_key = []
        image_key_control = []
        rotation_angle = []
        dataset_path = "/".join((self.output_entry, "instrument", "detector", "data"))
        n_dark = 0
        n_flat = 0
        n_proj = 0
        # handle darks
        try:
            with DatasetReader(self._data_darks_url) as dark_dataset:
                n_dark = dark_dataset.shape[0]
            # FIXME: avoid keeping some file open. not clear why this is needed
            dark_dataset = None
            FrameAppender(
                self._data_darks_url,
                self.output_file,
                data_path=dataset_path,
                where="end",
                logger=_logger,
            ).process()
        except Exception:
            _logger.error(
                f"No darks found in {self._data_darks_url.path()} or unable to add them to the dataset"
            )
        else:
            image_key.extend([ImageKey.DARK_FIELD.value] * n_dark)
            image_key_control.extend([ImageKey.DARK_FIELD.value] * n_dark)
            rotation_angle.extend([self.scan_range[0]] * n_dark)

        # handle flats
        try:
            with DatasetReader(self._data_flats_url) as flat_dataset:
                n_flat = flat_dataset.shape[0]
            # FIXME: avoid keeping some file open. not clear why this is needed
            flat_dataset = None
            FrameAppender(
                self._data_flats_url,
                self.output_file,
                data_path=dataset_path,
                where="end",
                logger=_logger,
            ).process()
        except Exception:
            _logger.warning(
                f"No flats found in {self._data_flats_url.path()} or unable to add them to the dataset"
            )
        else:
            image_key.extend([ImageKey.FLAT_FIELD.value] * n_flat)
            image_key_control.extend([ImageKey.FLAT_FIELD.value] * n_flat)
            rotation_angle.extend([self.scan_range[0]] * n_flat)

        # handle projections
        try:
            with DatasetReader(self._data_proj_url) as proj_dataset:
                n_proj = proj_dataset.shape[0]
            # FIXME: avoid keeping some file open. not clear why this is needed
            proj_dataset = None
            FrameAppender(
                self._data_proj_url,
                self.output_file,
                data_path=dataset_path,
                where="end",
                logger=_logger,
            ).process()
        except Exception:
            _logger.warning(
                f"No projections found in {self._data_proj_url.path()} or unable to add them to the dataset"
            )
        else:
            image_key.extend([ImageKey.PROJECTION.value] * n_proj)
            image_key_control.extend([ImageKey.PROJECTION.value] * n_proj)
            rotation_angle.extend(
                numpy.linspace(
                    self.scan_range[0],
                    self.scan_range[1],
                    num=n_proj,
                    endpoint=True,
                )
            )

        self._n_frames = n_flat + n_dark + n_proj

        with HDF5File(self.output_file, mode="a") as h5f:
            root_grp = h5f.require_group(self.output_entry)
            instrument_grp = root_grp.require_group("instrument")
            detector_grp = instrument_grp.require_group("detector")
            detector_grp["image_key"] = image_key
            detector_grp["image_key_control"] = image_key_control
            sample_grp = root_grp.require_group("sample")
            sample_grp["rotation_angle"] = rotation_angle
            sample_grp["rotation_angle"].attrs["unit"] = "degree"
            self._path_nxtomo_attrs(root_grp)

    def _read_sample_detector_distance(self):
        with EntryReader(self.input_root_url) as input_entry:
            path = "/".join(
                ("measurement", "instrument", "sample", "setup", "detector_distance")
            )
            dataset = input_entry[path]
            distance = h5py_read_dataset(dataset)
            unit = get_dataset_unit(
                dataset,
                default=_ureg.meter,
                from_dataset=f"{dataset.path} (reading sample-detector distance)",
            )
            return distance * unit

    def _read_energy(self):
        with EntryReader(self.input_root_url) as input_entry:
            path = "/".join(("measurement", "instrument", "source", "energy"))
            dataset = input_entry[path]
            energy = float(h5py_read_dataset(dataset))
            unit = get_dataset_unit(
                dataset=dataset,
                default=_ureg.keV,
                from_dataset=f"{dataset.name} (from reading energy)",
            )
            return energy * unit

    def _read_x_pixel_size(self):
        with EntryReader(self.input_root_url) as input_entry:
            path = "/".join(
                ("measurement", "instrument", "detector", "actual_pixel_size_x")
            )
            dataset = input_entry[path]
            pixel_size = float(h5py_read_dataset(dataset))
            unit = get_dataset_unit(
                dataset=dataset,
                default=_ureg.meter,
                from_dataset=f"{dataset.name} (from reading x pixel size)",
            )
            return pixel_size * unit

    def _read_y_pixel_size(self):
        with EntryReader(self.input_root_url) as input_entry:
            path = "/".join(
                ("measurement", "instrument", "detector", "actual_pixel_size_y")
            )
            dataset = input_entry[path]
            pixel_size = float(h5py_read_dataset(dataset))
            unit = get_dataset_unit(
                dataset=dataset,
                default=_ureg.meter,
                from_dataset=f"{dataset.name} (from reading y pixel size)",
            )
            return pixel_size * unit



from __future__ import annotations

from pydantic import Field, field_validator, ConfigDict
from collections import OrderedDict


from nxtomomill.models.utils import convert_str_to_tuple, filter_str_def
from ..base.general_section import GeneralSection as _GeneralSectionBase


class GeneralSection(_GeneralSectionBase):
    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    ewoksfluo_filename: str | None = Field(
        default=None,
        description="Path to the ewoksfluo-generated h5 file that contains fitted XRF data. If not provided from the configuration file must be provided from the command line.",
    )

    dimension: int = Field(
        default=3,
        description="Dimension of the experiment. 2 for 2D XRFCT, 3 for 3D XRFCT. Default is 3.",
    )

    detector_names: tuple[str, ...] = Field(
        default=tuple(),
        description="Define a list of (real or virtual) detector names used for the exp (space separated values - no comma). E.g. 'falcon xmap'. If not specified, all detectors are processed.",
    )

    @field_validator(
        "detector_names",
        mode="before",
    )
    @classmethod
    def cast_to_tuple(cls, value: str | None) -> tuple[str]:
        return convert_str_to_tuple(value) or ()

    @field_validator(
        "ewoksfluo_filename",
        mode="before",
    )
    @classmethod
    def cast_to_str(cls, value: str | None) -> str | None:
        return filter_str_def(value)

    def model_dump(self, *args, **kwargs) -> dict:
        unordered_result = super().model_dump(*args, **kwargs)
        ordered_result = OrderedDict(
            {
                "ewoksfluo_filename": unordered_result.pop("ewoksfluo_filename"),
                "output_file": unordered_result.pop("output_file"),
                "detector_names": unordered_result.pop("detector_names"),
                "dimension": unordered_result.pop("dimension"),
            }
        )
        ordered_result.update(unordered_result)
        return ordered_result



from __future__ import annotations

import numpy
import logging

from silx.io.url import DataUrl

from tomoscan.esrf.scan.utils import get_series_slice, get_n_series

from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey
from nxtomo.paths.nxtomo import get_paths as get_nexus_paths

from nxtomomill.converter.hdf5.acquisition.baseacquisition import BaseAcquisition
from nxtomomill.utils.utils import add_dark_flat_nx_file

_logger = logging.getLogger(__name__)

__all__ = [
    "ZSeriesDarkFlatCopy",
]


class ZSeriesDarkFlatCopy:
    """
    z-series version 3 can require to reuse dark-flat from a bliss scan done at the start or at the end of the series.

    This class is an helper to do this processing.
    """

    def __init__(
        self,
        series: list[BaseAcquisition],
        acquisition_to_nxtomo: dict[BaseAcquisition, tuple | None],
    ) -> None:
        """
        :param series: list of acquisition part of the series. **warning**: we expect this list to be ordered.
        :param acquisition_to_nxtomo: for each acquisition in "series" this dict can provide the nxtomo created during conversion. The value is expected to the None if the conversion failed or (file_path, data_path)

        .. warning:: even if list contains BaseAcquisition; those are sub_acquisitions of ZSeriesBaseAcquisition
        """
        for elmt in series:
            if not isinstance(elmt, BaseAcquisition):
                raise TypeError(
                    f"elmt is expected to be an instance of {BaseAcquisition}. Got {type(elmt)}"
                )

        self.series = series
        self.acquisition_to_nxtomo = acquisition_to_nxtomo

    @property
    def first_acquisition(self) -> BaseAcquisition:
        return self.series[0]

    @property
    def last_acquisition(self) -> BaseAcquisition:
        return self.series[-1]

    def run(self) -> None:
        darks_start_urls, flats_start_urls, darks_end_urls, flats_end_urls = (
            self.build_mapping_to_dark_flat_source()
        )
        edition_to_do = self.build_edition_to_do(
            darks_start_urls=darks_start_urls,
            flats_start_urls=flats_start_urls,
            darks_end_urls=darks_end_urls,
            flats_end_urls=flats_end_urls,
        )
        self.process_edition(editions_to_do=edition_to_do)

    def build_mapping_to_dark_flat_source(
        self,
    ) -> dict[tuple, dict[ImageKey, list[DataUrl]]]:
        """
        Build the list of DataUrl that can be used for copy.

        Output looks like:
        {
            file_path_first_nxtomo, data_path_first_nxtomo: {
                ImageKey.DARK_FIELD: (DataUrlSeries1, DataUrlSeries2),
                ImageKey.FLAT_FIELD: (DataUrlSeries1, ...),
            },
            file_path_last_nxtomo, data_path_last_nxtomo: {
                ImageKey.DARK_FIELD: (DataUrlSeries1, DataUrlSeries2),
                ImageKey.FLAT_FIELD: (DataUrlSeries1, ...),
            },
        }
        """
        darks_start_urls = []
        flats_start_urls = []
        darks_end_urls = []
        flats_end_urls = []
        # create darks_start_url and flats_start_url if possible
        first_nxtomos_infos = self.acquisition_to_nxtomo[self.first_acquisition][0]
        last_nxtomos_infos = self.acquisition_to_nxtomo[self.last_acquisition][-1]

        url_mapping = {}
        if first_nxtomos_infos is not None:
            url_mapping[first_nxtomos_infos] = {
                ImageKey.DARK_FIELD: darks_start_urls,
                ImageKey.FLAT_FIELD: flats_start_urls,
            }
        if last_nxtomos_infos is not None:
            url_mapping[last_nxtomos_infos] = {
                ImageKey.DARK_FIELD: darks_end_urls,
                ImageKey.FLAT_FIELD: flats_end_urls,
            }

        # build darks / flats urls
        for (file_path, data_path), image_key_to_urls in url_mapping.items():
            nxtomo_source = NXtomo().load(
                file_path=file_path,
                data_path=data_path,
            )
            for image_key, urls in image_key_to_urls.items():
                n_series = get_n_series(
                    image_key_values=nxtomo_source.instrument.detector.image_key_control,
                    image_key_type=image_key,
                )
                for i_series in range(n_series):
                    slice_source = get_series_slice(
                        image_key_values=nxtomo_source.instrument.detector.image_key_control,
                        image_key_type=image_key,
                        series_index=i_series,
                    )
                    if slice_source is None:
                        continue
                    detector_dataset_path = get_nexus_paths(version=None)
                    detector_data_path = "/".join(
                        (
                            data_path,
                            nxtomo_source.instrument.detector.path,
                            detector_dataset_path.nx_detector_paths.DATA,
                        )
                    )
                    urls.append(
                        DataUrl(
                            file_path=file_path,
                            data_path=detector_data_path,
                            data_slice=slice_source,
                            scheme="silx",
                        )
                    )

        return (
            tuple(darks_start_urls),
            tuple(flats_start_urls),
            tuple(darks_end_urls),
            tuple(flats_end_urls),
        )

    def build_edition_to_do(
        self,
        darks_start_urls: tuple[DataUrl],
        flats_start_urls: tuple[DataUrl],
        darks_end_urls: tuple[DataUrl],
        flats_end_urls: tuple[DataUrl],
    ) -> dict[tuple, dict]:
        """
        build all the edition to do to complete the dark and flat copy according to z-series v3 configuration.
        """
        # for each (final) acquisition register the different operation (edition) that needs to be done.
        editions_to_do: {tuple, dict} = {}
        # for each new entry (as a tuple of file_path, data_path) list the operations to execute

        for acquisition in self.series:
            if acquisition is None:
                # if conversion failed
                continue

            if acquisition._dark_at_start and (
                acquisition is not self.first_acquisition
            ):
                concatenate_dict(
                    editions_to_do,
                    {
                        nx_tomo: {"darks_start": darks_start_urls}
                        for nx_tomo in self.acquisition_to_nxtomo[acquisition]
                    },
                )
            if acquisition._flat_at_start and (
                acquisition is not self.first_acquisition
            ):
                concatenate_dict(
                    editions_to_do,
                    {
                        nx_tomo: {"flats_start": flats_start_urls}
                        for nx_tomo in self.acquisition_to_nxtomo[acquisition]
                    },
                )
            if acquisition._dark_at_end and (acquisition is not self.last_acquisition):
                concatenate_dict(
                    editions_to_do,
                    {
                        nx_tomo: {"darks_end": darks_end_urls}
                        for nx_tomo in self.acquisition_to_nxtomo[acquisition]
                    },
                )
            if acquisition._flat_at_end and (acquisition is not self.last_acquisition):
                concatenate_dict(
                    editions_to_do,
                    {
                        nx_tomo: {"flats_end": flats_end_urls}
                        for nx_tomo in self.acquisition_to_nxtomo[acquisition]
                    },
                )
        return editions_to_do

    def process_edition(
        self, editions_to_do: dict[tuple, dict], embed_dark_flat: bool = True
    ):
        """
        do edition
        """
        for nxtomo_to_edit, params_to_urls in editions_to_do.items():
            for param_name, urls in params_to_urls.items():
                edited_nxtomo_file_path, edited_nxtomo_data_path = nxtomo_to_edit
                for url in urls:
                    # if the file already contains the dark / flat
                    if (
                        url.file_path() == edited_nxtomo_file_path
                        and url.data_path() == edited_nxtomo_data_path
                    ):
                        continue
                    add_dark_flat_nx_file(
                        **{
                            param_name: url,
                            "file_path": edited_nxtomo_file_path,
                            "entry": edited_nxtomo_data_path,
                            "embed_data": embed_dark_flat,
                            "logger": _logger,
                        }
                    )


def concatenate_dict(dict_1: dict, dict_2: dict) -> None:
    """
    concatenate two dicts into dict_1
    """
    assert isinstance(dict_1, dict)
    assert isinstance(dict_2, dict)
    for key in dict_2.keys():
        if key in dict_1.keys():
            if isinstance(dict_1[key], dict) and isinstance(dict_2[key], dict):
                concatenate_dict(dict_1=dict_1[key], dict_2=dict_2[key])
            else:
                dict_1[key] = numpy.concatenate((dict_1[key], dict_2[key]))
        else:
            dict_1[key] = dict_2[key]



# coding: utf-8

import os
import tempfile

from tomoscan.esrf.mock import MockEDF

from nxtomomill.app.edf2nx import main


def test_edf2nx_application():
    """test nxtomomill edf2nx input_file output_file"""
    with tempfile.TemporaryDirectory() as folder:
        edf_acq_path = os.path.join(folder, "acquisition")
        n_proj = 10
        MockEDF(
            scan_path=edf_acq_path,
            n_radio=n_proj,
            n_ini_radio=n_proj,
        )
        output_file = os.path.join(edf_acq_path, "nexus_file.nx")

        assert not os.path.exists(output_file), "output_file exists already"
        main(["edf2nx", edf_acq_path, output_file])
        assert os.path.exists(output_file), "output_file doesn't exists"



from __future__ import annotations

import os
from tomoscan.utils.io import filter_esrf_mounting_points


def abspath(path: str):
    """
    File absolute path, removing 'esrf' mounting point.
    Those mounting point must be remove because they might be differente from one computer to the other
    """
    return filter_esrf_mounting_points(os.path.abspath(path))



import sys
import logging
import argparse
from nxtomomill.converter.fscan.fscanconverter import from_fscan_to_nx

_logger = logging.getLogger(__name__)


def main(argv):
    parser = argparse.ArgumentParser(
        description="convert fscan dataset to nexus format. "
    )
    parser.add_argument(
        "input_file",
        help="Path to the master file of the acquisition, containing entries '1.1', '2.1', etc",
        default=None,
        nargs="?",
    )
    parser.add_argument(
        "output_file",
        help="Path to the output NX file",
        default=None,
        nargs="?",
    )
    parser.add_argument(
        "--halftomo",
        help="Whether to enable extended-field-of-view mode (half-tomography). Default is False.",
        dest="halftomo",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--ignore_last_n_projections",
        "--ignore-last-n-projections",
        default=0,
        help="Number of projections to ignore in the end of each series of projections",
    )

    parser.add_argument(
        "--detector-name",
        "--detector_name",
        default="pcoedgehs",
        help="Detector name. Default is pcoedgehs.",
    )
    parser.add_argument(
        "--rot_motor_name",
        "--rot-motor-name",
        default="hrrz",
        help="Rotation motor name. Default is hrrz.",
    )

    parser.add_argument("--energy", default=None, help="Incident beam energy in keV")
    parser.add_argument(
        "--distance", default=None, help="Sample-detector distance in meters"
    )

    options = parser.parse_args(argv[1:])

    if options.input_file is None or options.output_file is None:
        _logger.error("Need input file and output file. Type --help to see options")
        sys.exit(0)

    energy = options.energy
    distance = options.distance
    if energy is not None:
        energy = float(energy)
    if distance is not None:
        distance = float(distance)

    from_fscan_to_nx(
        options.input_file,
        options.output_file,
        detector_name=options.detector_name,
        rotation_motor_name=options.rot_motor_name,
        halftomo=options.halftomo,
        ignore_last_n_projections=options.ignore_last_n_projections,
        energy=energy,
        distance=distance,
    )


if __name__ == "__main__":
    main(sys.argv)



from __future__ import annotations

from nxtomomill.utils.io import deprecated
from .ConfigBase import ConfigBase


class FrmFlatToNestedBase(ConfigBase):
    """
    Base class defining behavior for models needing to be dumped with sections.

    Internally they are represented as 'flat' but externally they are represented as nested to ease user
    setting values.
    This is the historical design. Maybe in the future we will use nested models internally as well.
    """

    def to_cfg_file(self, file_path: str, filter_sections: tuple[str] = tuple()):
        nested_model = self.to_nested_model()
        nested_model.to_cfg_file(
            file_path=file_path,
            filter_sections=filter_sections,
        )

    def to_nested_model(self):
        raise NotImplementedError

    @deprecated(
        reason="Replaced by pydantic function 'model_dump'", since_version="2.0"
    )
    def to_dict(self) -> dict:
        config = self.to_nested_model()
        return {key.upper(): value for key, value in config.model_dump().items()}



from __future__ import annotations

from pydantic import Field, field_serializer, field_validator, ConfigDict

from nxtomo.nxobject.nxdetector import FieldOfView

from nxtomomill.models.utils import filter_str_def, convert_str_to_bool
from ..base.general_section import GeneralSection as _GeneralSectionBase


class GeneralSection(_GeneralSectionBase):
    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    input_file: str | None = Field(default=None, description="Input file path.")
    raises_error: bool = Field(
        default=False, description="Raise an error when encountered."
    )
    no_input: bool = Field(
        default=False, description="Ask for user inputs if missing information."
    )
    single_file: bool = Field(
        default=False, description="Create a single file for all sequences."
    )
    no_master_file: bool = Field(
        default=True, description="Avoid creating the master file."
    )
    ignore_bliss_tomo_config: bool = Field(
        default=False,
        description="Ignore Bliss tomography group. On recent bliss file (2023) a dedicated group specify datasets to be used for tomography. Defining for example translations, rotation, etc. If True then this group will be ignored and conversion will fallback on using path list provided in the KEYS section.",
    )
    field_of_view: FieldOfView | None = Field(
        default=None,
        description="Force output to be a 'Full' or 'Half' acquisition. Else determine from existing metadata",
    )
    create_control_data: bool = Field(
        default=True, description="Generate control/data - filled with machine current."
    )
    check_tomo_n: bool | None = Field(
        default=None,
        description="Check 'tomo_n' metadata once the conversion is done. By default will be done for all except scan build from 'scan_data'.",
    )
    rotation_is_clockwise: bool | None = Field(
        default=None,
        description="Force rotation clockwise or not. If not set will read the information from the bliss-tomo 'tomoconfig' group. Else consider rotation is counterclockwise.",
    )
    mechanical_lr_flip: bool = Field(
        default=False,
        description="Detector image is flipped **left-right** for mechanical reasons that are not propagated with bliss-tomo hdf5 metadata (mirror,..).",
    )
    mechanical_ud_flip: bool = Field(
        default=False,
        description="Detector image is flipped **up-down** for mechanical reasons that are not propagated with bliss-tomo hdf5 metadata (mirror,..).",
    )

    @property
    def request_input(self) -> bool:
        return not self.no_input

    @request_input.setter
    def request_input(self, request: bool):
        self.no_input = request

    @field_validator(
        "check_tomo_n",
        "rotation_is_clockwise",
        mode="before",
    )
    @classmethod
    def cast_general_section_to_bool_or_none(
        cls, value: str | None | bool
    ) -> bool | None:
        if isinstance(value, str):
            value = filter_str_def(value)
        if value in (None, "None", ""):
            return None
        elif value in ("True", True, "1"):
            return True
        elif value in ("False", False, "0"):
            return False
        else:
            raise ValueError

    @field_serializer(
        "check_tomo_n",
        "rotation_is_clockwise",
        when_used="always",
    )
    @classmethod
    def serialize_check_tomo_n(cls, value: None | bool) -> str:
        if value is None:
            return None
        else:
            return value

    @field_validator(
        "field_of_view",
        mode="plain",
    )
    @classmethod
    def cast_to_field_of_view(cls, value: str | FieldOfView) -> FieldOfView | None:
        if value in (None, ""):
            return None
        return FieldOfView(value)

    @field_serializer(
        "field_of_view",
        when_used="always",
    )
    @classmethod
    def serialize_field_of_view(cls, field_of_view: FieldOfView | None) -> str:
        if field_of_view is None:
            return None
        else:
            return field_of_view.value

    @field_validator(
        "input_file",
        mode="before",
    )
    @classmethod
    def cast_general_section_to_str(cls, value: str | None) -> str | None:
        return filter_str_def(value)

    @field_validator(
        "no_input",
        "single_file",
        "no_master_file",
        "ignore_bliss_tomo_config",
        "create_control_data",
        "mechanical_lr_flip",
        "mechanical_ud_flip",
        mode="before",
    )
    @classmethod
    def cast_general_section_to_bool(cls, value: str | bool | None) -> bool:
        return convert_str_to_bool(value)



# coding: utf-8


import unittest

from silx.io.url import DataUrl

from nxtomomill.models.h52nx._acquisitionstep import AcquisitionStep
from nxtomomill.models.h52nx.FrameGroup import (
    FrameGroup,
    filter_acqui_frame_type,
)


class TestFrameGroupCreator(unittest.TestCase):
    """Test FrameGroup class"""

    def setUp(self) -> None:
        self.url1 = DataUrl(
            file_path="/path/to/file/my_file", data_path="/path/to/data", scheme="silx"
        )
        self.url2 = DataUrl(file_path="my_file", data_path="/path/to/data")

    def test_default_constructor(self):
        """Test FrameGroup constructor"""
        frame_grp = FrameGroup(url=self.url1, frame_type="projection")
        self.assertEqual(frame_grp.url.path(), self.url1.path())
        self.assertEqual(frame_grp.frame_type, AcquisitionStep.PROJECTION)

        frame_grp = FrameGroup(url=self.url2, frame_type=AcquisitionStep.ALIGNMENT)
        self.assertEqual(frame_grp.url.path(), self.url2.path())
        self.assertEqual(frame_grp.frame_type, AcquisitionStep.ALIGNMENT)

        with self.assertRaises(ValueError):
            FrameGroup(url=self.url1, frame_type="toto")

        with self.assertRaises(ValueError):
            FrameGroup(url=self.url1, frame_type="projection", copy_data="tata")

    def test_constructor_frm_str(self):
        """Test FrameGroup.frm_str function"""
        grp_prefix = FrameGroup.frm_str(
            "frame_type=projections, "
            "entry=silx:///path/to/file/my_file?/path/to/data, "
            "copy_data=True"
        )
        self.assertTrue(isinstance(grp_prefix, FrameGroup))
        self.assertEqual(grp_prefix.copy_data, True)

        grp_no_prefix = FrameGroup.frm_str(
            "projections, silx:///path/to/file/my_file?/path/to/data, True"
        )
        self.assertEqual(str(grp_no_prefix), str(grp_prefix))

        with self.assertRaises(ValueError):
            FrameGroup.frm_str(
                "frame_type=projections, "
                "entry=silx:///path/to/file/my_file?/path/to/data, "
                "copy_data=12"
            )

        with self.assertRaises(ValueError):
            FrameGroup.frm_str("frame_type=projections")


class TestFilterCurrentAcquiFrameType(unittest.TestCase):
    """test filter_acqui_frame_type function"""

    def setUp(self) -> None:
        self.init_1 = FrameGroup(frame_type="init", url=None)
        self.sequence1 = (
            self.init_1,
            FrameGroup(frame_type="dark", url="data/to/dark/1"),
            FrameGroup(frame_type="flat", url="data/to/flat/1"),
            FrameGroup(frame_type="proj", url="data/to/proj/1"),
            FrameGroup(frame_type="proj", url="data/to/proj/2"),
        )

        self.init_2 = FrameGroup(frame_type="init", url="/path/to/init")
        self.sequence2 = (
            self.init_2,
            FrameGroup(frame_type="dark", url="data/to/dark/2"),
            FrameGroup(frame_type="proj", url="data/to/proj/3"),
            FrameGroup(frame_type="proj", url="data/to/proj/4"),
            FrameGroup(frame_type="proj", url="data/to/proj/5"),
            FrameGroup(frame_type="proj", url="data/to/proj/6"),
            FrameGroup(frame_type="flat", url="data/to/flat/2"),
            FrameGroup(frame_type="alignment", url="data/to/alignment/1"),
        )

    def test_search_init(self):
        """test filter_acqui_frame_type with init frame group"""
        with self.assertRaises(ValueError):
            filter_acqui_frame_type(
                init=self.init_1,
                sequences=self.sequence1,
                frame_type=AcquisitionStep.INITIALIZATION,
            )

    def test_search_flat(self):
        """test filter_acqui_frame_type with flat frame group"""
        flat_url = filter_acqui_frame_type(
            init=self.init_1,
            sequences=self.sequence1,
            frame_type=AcquisitionStep.FLAT,
        )
        self.assertEqual(len(flat_url), 1)
        self.assertEqual(flat_url[0].url.path(), self.sequence1[2].url.path())

        with self.assertRaises(ValueError):
            filter_acqui_frame_type(
                init=self.init_1,
                sequences=self.sequence2,
                frame_type=AcquisitionStep.FLAT,
            )

    def test_search_dark(self):
        """test filter_acqui_frame_type with dark frame group"""
        seq_sum = list(self.sequence1)
        seq_sum.extend(self.sequence2)
        dark_1 = filter_acqui_frame_type(
            init=self.init_1, sequences=tuple(seq_sum), frame_type=AcquisitionStep.DARK
        )
        self.assertEqual(len(dark_1), 1)
        self.assertEqual(dark_1[0].url.path(), self.sequence1[1].url.path())
        dark_2 = filter_acqui_frame_type(
            init=self.init_2, sequences=tuple(seq_sum), frame_type=AcquisitionStep.DARK
        )
        self.assertEqual(len(dark_2), 1)
        self.assertEqual(dark_2[0].url.path(), self.sequence2[1].url.path())

    def test_search_proj(self):
        """test filter_acqui_frame_type with projection frame group"""
        projs_1 = filter_acqui_frame_type(
            init=self.init_2,
            sequences=self.sequence2,
            frame_type=AcquisitionStep.PROJECTION,
        )
        self.assertEqual(len(projs_1), 4)
        self.assertEqual(projs_1[2].url.path(), self.sequence2[4].url.path())

        seq_sum = list(self.sequence1)
        seq_sum.extend(self.sequence2)
        projs_2 = filter_acqui_frame_type(
            init=self.init_2,
            sequences=tuple(seq_sum),
            frame_type=AcquisitionStep.PROJECTION,
        )
        self.assertEqual(len(projs_2), 4)
        self.assertEqual(projs_1[1].url.path(), projs_2[1].url.path())



# coding: utf-8

import os
import numpy
import pytest

from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey

from tomoscan.io import HDF5File, get_swmr_mode
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan

from nxtomomill import converter
from nxtomomill.converter.hdf5.acquisition.multitomo import MultiTomoAcquisition
from nxtomomill.io.config import TomoHDF5Config
from nxtomomill.tests.utils.bliss import MockBlissAcquisition

configs = (
    {"nb_tomo": 1, "nb_loop": 5, "nb_turns": 5},
    {"nb_tomo": 5, "nb_loop": 1, "nb_turns": 5},
    {"nb_tomo": 2, "nb_loop": 3, "nb_turns": 6},
)


@pytest.mark.parametrize("multitomo_version", (1, 2))
@pytest.mark.parametrize("config", configs)
def test_multitomo_conversion(tmp_path, config: dict, multitomo_version: int):
    n_darks = 10
    n_flats = 10
    bliss_scan_dir = str(tmp_path / "my_acquisition")
    nb_tomo = config["nb_tomo"]
    nb_loop = config["nb_loop"]
    nb_turns = config["nb_turns"]
    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=1,
        n_darks=n_darks,
        n_flats=n_flats,
        output_dir=bliss_scan_dir,
        acqui_type="multitomo",
        with_rotation_motor_info=True,
        nb_tomo=nb_tomo if multitomo_version == 1 else None,
        nb_loop=nb_loop if multitomo_version == 1 else None,
        nb_turns=nb_turns if multitomo_version == 2 else None,
    )
    sample_file = bliss_mock.samples[0].sample_file

    configuration = TomoHDF5Config()

    output_nx_file = os.path.join(str(tmp_path), "nexus_scans.nx")
    assert not os.path.exists(output_nx_file)
    configuration.output_file = output_nx_file
    configuration.input_file = sample_file
    configuration.no_input = True
    configuration.raises_error = False
    configuration.no_master_file = False

    result = converter.from_h5_to_nx(configuration=configuration)
    assert os.path.exists(output_nx_file)
    # check we have our 5 NXtomo converted
    assert len(result) == nb_loop * nb_tomo
    n_projs = 10 * nb_loop * nb_tomo
    # check saved datasets
    with HDF5File(sample_file, mode="r", swmr=get_swmr_mode()) as h5f:
        darks = h5f["2.1/instrument/pcolinux/data"][()]
        assert len(darks) == n_darks
        flats_1 = h5f["3.1/instrument/pcolinux/data"][()]
        assert len(flats_1) == n_flats
        projections = h5f["4.1/instrument/pcolinux/data"][()]
        assert len(projections) == n_projs
        flats_2 = h5f["5.1/instrument/pcolinux/data"][()]
        assert len(flats_2) == n_flats

    for i_nx_tomo, (file_path, data_path) in enumerate(result):
        with HDF5File(file_path, mode="r", swmr=get_swmr_mode()) as h5f:
            detector_path = "/".join([data_path, "instrument", "detector", "data"])
            detector_data = h5f[detector_path][()]
            expected_data = numpy.concatenate(
                [
                    darks,
                    flats_1,
                    projections[(10 * i_nx_tomo) : (10 * (i_nx_tomo + 1))],
                    flats_2,
                ]
            )
            numpy.testing.assert_array_equal(detector_data, expected_data)
            sequence_number_path = "/".join(
                [data_path, "instrument", "detector", "sequence_number"]
            )
            sequence_number = h5f[sequence_number_path]
            numpy.testing.assert_array_equal(
                sequence_number,
                numpy.concatenate(
                    (
                        numpy.linspace(
                            start=0,
                            stop=n_darks,
                            num=n_darks,
                            endpoint=False,
                            dtype=numpy.uint32,
                        ),  # darks
                        numpy.linspace(
                            start=n_darks,
                            stop=n_darks + n_flats,
                            num=n_flats,
                            endpoint=False,
                            dtype=numpy.uint32,
                        ),  # flats 1
                        numpy.linspace(
                            start=i_nx_tomo * 10 + n_darks + n_flats,
                            stop=(i_nx_tomo + 1) * 10 + n_darks + n_flats,
                            num=10,
                            endpoint=False,
                            dtype=numpy.uint32,
                        ),
                        numpy.linspace(
                            start=n_darks + n_flats + n_projs,
                            stop=n_darks + 2 * n_flats + n_projs,
                            num=n_flats,
                            endpoint=False,
                            dtype=numpy.uint32,
                        ),  # flats 2
                    )
                ),
            )


def test_multitomo_conversion_with_angle_subselection(tmp_path):
    bliss_scan_dir = str(tmp_path / "my_acquisition")
    nb_tomo = 1
    nb_loop = 1
    bliss_mock = MockBlissAcquisition(
        n_sample=1,
        n_sequence=1,
        n_scan_per_sequence=1,
        n_darks=1,
        n_flats=1,
        output_dir=bliss_scan_dir,
        acqui_type="multitomo",
        with_rotation_motor_info=True,
        nb_tomo=nb_tomo,
        nb_loop=nb_loop,
    )
    sample_file = bliss_mock.samples[0].sample_file

    configuration = TomoHDF5Config()

    output_nx_file = os.path.join(str(tmp_path), "nexus_scans.nx")
    assert not os.path.exists(output_nx_file)
    configuration.output_file = output_nx_file
    configuration.input_file = sample_file
    configuration.no_input = True
    configuration.raises_error = False
    configuration.multitomo_scan_range = 90
    configuration.multitomo_start_angle_offset = 10
    configuration.multitomo_n_nxtomo = 2
    configuration.raises_error = True
    configuration.multitomo_shift_angles = False

    file_path, data_path = converter.from_h5_to_nx(configuration=configuration)[0]
    scan = NXtomoScan(file_path, data_path)
    is_proj_mask = scan.image_key == ImageKey.PROJECTION.value
    remaining_proj_angle = numpy.array(scan.rotation_angle)[is_proj_mask]
    assert remaining_proj_angle.min() >= 10
    assert remaining_proj_angle.max() <= 100

    configuration.multitomo_shift_angles = True
    configuration.overwrite = True
    file_path, data_path = converter.from_h5_to_nx(configuration=configuration)[0]
    scan = NXtomoScan(file_path, data_path)
    is_proj_mask = scan.image_key == ImageKey.PROJECTION.value
    remaining_proj_angle = numpy.array(scan.rotation_angle)[is_proj_mask]
    assert remaining_proj_angle.min() >= 0
    assert remaining_proj_angle.max() <= 90


def test_get_projections_slices():
    """test the _get_projections_slices function"""
    nx_tomo = NXtomo()
    nx_tomo.instrument.detector.image_key_control = [2, 1, 0, 0, 0, 2]
    assert MultiTomoAcquisition._get_projections_slices(nx_tomo) == (slice(2, 5, 1),)
    nx_tomo.instrument.detector.image_key_control = [2, 1, 0, 0, 0, 2]
    assert MultiTomoAcquisition._get_projections_slices(nx_tomo) == (slice(2, 5, 1),)
    nx_tomo.instrument.detector.image_key_control = [2, 1]
    assert MultiTomoAcquisition._get_projections_slices(nx_tomo) == ()
    nx_tomo.instrument.detector.image_key_control = [0, 0, 2, 1, 0, 0, 1, 0, 0]
    assert MultiTomoAcquisition._get_projections_slices(nx_tomo) == (
        slice(0, 2, 1),
        slice(4, 6, 1),
        slice(7, 9, 1),
    )



from __future__ import annotations

import pint
from nxtomo.nxobject.nxdetector import FieldOfView
from nxtomomill.utils.io import deprecated

from pydantic import BaseModel, Field, field_validator, field_serializer, ConfigDict

_ureg = pint.get_application_registry()


class UnitSection(BaseModel):

    model_config = ConfigDict(
        validate_assignment=True, arbitrary_types_allowed=True, validate_by_name=True
    )

    pixel_size_unit: pint.Unit = Field(
        default_factory=lambda: _ureg.micrometer,
        description="Unit used to save pixel size. Must be in of ['nanometer', 'micrometer', 'millimeter', 'centimeter', 'meter']",
        alias="expected_unit_for_pixel_size",
    )
    sample_detector_distance_unit: pint.Unit | None = Field(
        default_factory=lambda: _ureg.millimeter,
        description="Unit used by SPEC to save sample to detector distance. Must be in of ['nanometer', 'micrometer', 'millimeter', 'centimeter', 'meter']",
        alias="expected_unit_for_distance",
    )
    energy_unit: pint.Unit | None = Field(
        default_factory=lambda: _ureg.keV,
        description="Unit used by SPEC to save energy. Must be in of ['joule', 'electron_volt', 'kiloelectron_volt', 'millielectron_volt', 'gigaelectron_volt', 'kilojoule']",
        alias="expected_unit_for_energy",
    )
    x_translation_unit: pint.Unit | None = Field(
        default_factory=lambda: _ureg.millimeter,
        description="Unit used by SPEC to save x translation. Must be in of ['nanometer', 'micrometer', 'millimeter', 'centimeter', 'meter']",
        alias="expected_unit_for_x_translation",
    )
    y_translation_unit: pint.Unit | None = Field(
        default_factory=lambda: _ureg.millimeter,
        description="Unit used by SPEC to save y translation. Must be in of ['nanometer', 'micrometer', 'millimeter', 'centimeter', 'meter']",
        alias="expected_unit_for_y_translation",
    )
    z_translation_unit: pint.Unit | None = Field(
        default_factory=lambda: _ureg.millimeter,
        description="Unit used by SPEC to save z translation. Must be in of ['nanometer', 'micrometer', 'millimeter', 'centimeter', 'meter']",
        alias="expected_unit_for_z_translation",
    )
    machine_current_unit: pint.Unit | None = Field(
        default_factory=lambda: _ureg.mA,
        description="Unit used by SPEC to save machine current (also aka SRcurrent). Must be in of ['ampere', 'kiloampere', 'milliampere']",
        alias="expected_unit_for_machine_current",
    )

    @field_validator(
        "pixel_size_unit",
        "sample_detector_distance_unit",
        "energy_unit",
        "x_translation_unit",
        "y_translation_unit",
        "z_translation_unit",
        "machine_current_unit",
        mode="before",
    )
    @classmethod
    def cast_to_pint_unit(cls, value: str | FieldOfView) -> pint.Unit | None:
        if value in ("kiloelectron_volt", "kev"):
            return _ureg.keV
        if value == "milliampere":
            return _ureg.mA
        return pint.Unit(value)

    @field_serializer(
        "pixel_size_unit",
        "sample_detector_distance_unit",
        "energy_unit",
        "x_translation_unit",
        "y_translation_unit",
        "z_translation_unit",
        "machine_current_unit",
        when_used="always",
    )
    @classmethod
    def serialized_unit(cls, value) -> str:
        return f"{value}"

    @property
    @deprecated(
        reason="renamed",
        replacement="sample_detector_distance_unit",
        since_version="2.0",
    )
    def distance_unit(self):
        return self.sample_detector_distance_unit

    @distance_unit.setter
    @deprecated(
        reason="renamed",
        replacement="sample_detector_distance_unit",
        since_version="2.0",
    )
    def distance_unit(self, unit):
        self.sample_detector_distance_unit = unit


UnitSection.model_rebuild()



from __future__ import annotations

from nxtomo.nxobject.nxdetector import FieldOfView
from nxtomomill.models.utils import filter_str_def

from pydantic import BaseModel, Field, field_validator, ConfigDict, field_serializer


class DetectorSection(BaseModel):

    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    field_of_view: FieldOfView | None = Field(
        default=None,
        description="Detector field of view. If set must be in `Half` or `Full`",
    )

    @field_validator(
        "field_of_view",
        mode="plain",
    )
    @classmethod
    def cast_to_field_of_view(cls, value: str | FieldOfView) -> FieldOfView | None:
        if value in (None, ""):
            return None
        if isinstance(value, str):
            value = filter_str_def(value)
        return FieldOfView(value)

    @field_serializer(
        "field_of_view",
        when_used="always",
    )
    @classmethod
    def serialize_field_of_view(cls, field_of_view: FieldOfView | None) -> str:
        if field_of_view is None:
            return None
        else:
            return field_of_view.value
